unit Base64;

interface

uses
  Classes, SysUtils;

function GetBase64EncodeSize(const InputSize: Integer): Integer;
function GetBase64DecodeSize(const InputSize: Integer): Integer;

procedure Base64Encode(Input: Pointer; InputSize: Integer; Output: Pointer; var OutputSize: Integer);
procedure Base64Decode(Input: Pointer; InputSize: Integer; Output: Pointer; var OutputSize: Integer);

function Base64EncodeAnsiString(const Input: AnsiString): AnsiString;
function Base64DecodeAnsiString(const Input: AnsiString): AnsiString;

function Base64EncodeString(const Input: string; const AEncoding: TEncoding = nil): string;
function Base64DecodeString(const Input: string; const AEncoding: TEncoding = nil): string;

procedure Base64EncodeStream(const InputStream: TStream; const OutputStream: TStream);
procedure Base64DecodeStream(const InputStream: TStream; const OutputStream: TStream);

function AddLineBreak(Line: string; SpaceLen: Integer = 4; LineLen: Integer = 80): string;
function TrimIllegalChar(Line: AnsiString): AnsiString; overload;
function TrimIllegalChar(Line: string): string; overload;

implementation

{$Q-}{$R-}

function TrimIllegalChar(Line: AnsiString): AnsiString;
var
  i, l: Integer;
begin
  l  := 1;
  SetLength(Result, Length(Line));
  for i  := 1 to Length(Line) do
    if (Line[i] in ['A'..'Z']) or (Line[i] in ['a'..'z']) or
        (Line[i] in ['0'..'9']) or (Line[i] in ['+', '/', '=']) then
    begin
      Result[l]  := Line[i];
      Inc(l);
    end;
  SetLength(Result, l-1);
end;

function TrimIllegalChar(Line: string): string;
var
  i, l: Integer;
begin
  l  := 1;
  SetLength(Result, Length(Line));
  for i  := 1 to Length(Line) do
    if CharInSet(Line[i], ['A'..'Z']) or CharInSet(Line[i], ['a'..'z']) or
        CharInSet(Line[i], ['0'..'9']) or CharInSet(Line[i], ['+', '/', '=']) then
    begin
      Result[l]  := Line[i];
      Inc(l);
    end;
  SetLength(Result, l-1);
end;

function AddLineBreak(Line: string; SpaceLen: Integer = 4; LineLen: Integer = 80): string;
var
  Space: string;
begin
  Result := '';
  Space := StringOfChar(' ', SpaceLen);
  while Line <> '' do
  begin
    Result := Result + Space + Copy(Line, 1, LineLen) + #13#10;
    Delete(Line, 1, LineLen);
  end;
end;

const
  B64: array[0..63] of byte= (65,66,67,68,69,70,71,72,73,74,75,76,77,78,79,80,
    81,82,83,84,85,86,87,88,89,90,97,98,99,100,101,102,103,104,105,106,107,108,
    109,110,111,112,113,114,115,116,117,118,119,120,121,122,48,49,50,51,52,53,
    54,55,56,57,43,47);

function GetBase64EncodeSize(const InputSize: Integer): Integer;
begin
  Result := ((InputSize+2) div 3) * 4;
end;

function GetBase64DecodeSize(const InputSize: Integer): Integer;
begin
  Result := (InputSize div 4) * 3;
end;

procedure Base64Encode(Input: Pointer; InputSize: Integer; Output: Pointer; var OutputSize: Integer);
var
  i, ip, op: integer;
  InputBA, OutputBA: PByteArray;
begin
  InputBA := PByteArray(Input);
  OutputBA := PByteArray(Output);

  ip := 0;
  op := 0;

  for i := 1 to (InputSize div 3) do
  begin
    OutputBA^[op+0] := B64[InputBA^[ip] shr 2];
    OutputBA^[op+1] := B64[((InputBA^[ip] and 3) shl 4) + (InputBA^[ip+1] shr 4)];
    OutputBA^[op+2] := B64[((InputBA^[ip+1] and 15) shl 2) + (InputBA^[ip+2] shr 6)];
    OutputBA^[op+3] := B64[InputBA^[ip+2] and 63];
    Inc(op,4); Inc(ip,3);
  end;

  case (InputSize mod 3) of
    1: begin
         OutputBA^[op+0] := B64[InputBA^[ip] shr 2];
         OutputBA^[op+1] := B64[(InputBA^[ip] and 3) shl 4];
         OutputBA^[op+2] := byte('=');
         OutputBA^[op+3] := byte('=');
       end;
    2: begin
         OutputBA^[op+0] := B64[InputBA^[ip] shr 2];
         OutputBA^[op+1] := B64[((InputBA^[ip] and 3) shl 4) + (InputBA^[ip+1] shr 4)];
         OutputBA^[op+2] := B64[(InputBA^[ip+1] and 15) shl 2];
         OutputBA^[op+3] := byte('=');
       end;
  end;

  OutputSize := ((InputSize+2) div 3) * 4;
end;

procedure Base64Decode(Input: Pointer; InputSize: Integer; Output: Pointer; var OutputSize: Integer);
var
  i, j, ip, op: integer;
  Temp: array[0..3] of byte;
  InputBA, OutputBA: PByteArray;
begin
  InputBA := PByteArray(Input);
  OutputBA := PByteArray(Output);

  ip := 0;
  op := 0;

  OutputSize := 0;
  for i := 1 to (InputSize div 4) do
  begin
    for j := 0 to 3 do
    begin
      case InputBA^[ip] of
        65..90 : Temp[j] := InputBA^[ip] - Ord('A');
        97..122: Temp[j] := InputBA^[ip] - Ord('a') + 26;
        48..57 : Temp[j] := InputBA^[ip] - Ord('0') + 52;
        43     : Temp[j] := 62;
        47     : Temp[j] := 63;
        61     : Temp[j] := $FF;
      end;
      Inc(ip);
    end;
    OutputBA^[op] := (Temp[0] shl 2) or (Temp[1] shr 4);
    OutputSize := op+1;
    if (Temp[2]<> $FF) and (Temp[3]= $FF) then
    begin
      OutputBA^[op+1] := (Temp[1] shl 4) or (Temp[2] shr 2);
      OutputSize := op+2;
      Inc(op)
    end
    else if (Temp[2]<> $FF) then
    begin
      OutputBA^[op+1] := (Temp[1] shl 4) or (Temp[2] shr 2);
      OutputBA^[op+2] := (Temp[2] shl 6) or  Temp[3];
      OutputSize := op+3;
      Inc(op,2);
    end;
    Inc(op);
  end;
end;

function Base64EncodeAnsiString(const Input: AnsiString): AnsiString;
var
  ResultSize: Integer;
begin
  SetLength(Result, ((Length(Input)+2) div 3) * 4);
  Base64Encode(@Input[1], Length(Input), @Result[1], ResultSize);
end;

function Base64DecodeAnsiString(const Input: AnsiString): AnsiString;
var
  ResultSize: Integer;
  TrimInput: AnsiString;
begin
  TrimInput := TrimIllegalChar(Input);
  SetLength(Result, (Length(TrimInput) div 4) * 3);
  Base64Decode(@TrimInput[1], Length(TrimInput), @Result[1], ResultSize);
  SetLength(Result, ResultSize);
end;

function Base64EncodeString(const Input: string; const AEncoding: TEncoding = nil): string;
var
  BinData: TBytes;
  RetData: AnsiString;
  ResultSize: Integer;
begin
  if Assigned(AEncoding) then
    BinData := AEncoding.GetBytes(Input)
  else
    BinData := TEncoding.Default.GetBytes(Input);
  SetLength(RetData, ((Length(BinData)+2) div 3) * 4);
  Base64Encode(@BinData[0], Length(BinData), @RetData[1], ResultSize);
  Result := string(RetData);
end;

function Base64DecodeString(const Input: string; const AEncoding: TEncoding = nil): string;
var
  TrimInput: string;
  ResultSize: Integer;
  BinData: TBytes;
begin
  TrimInput := TrimIllegalChar(Input);
  SetLength(BinData, (Length(TrimInput) div 4) * 3);
  Base64Decode(@TrimInput[1], Length(TrimInput), @BinData[0], ResultSize);
  SetLength(BinData, ResultSize);

  if Assigned(AEncoding) then
    Result := AEncoding.GetString(BinData)
  else
    Result := TEncoding.Default.GetString(BinData);
end;

procedure Base64EncodeStream(const InputStream: TStream; const OutputStream: TStream);
var
  InputBuf: PAnsiChar;
  InputBufSize: Integer;
  OutputBuf: PAnsiChar;
  OutputBufSize: Integer;
begin
  InputBuf := AllocMem(4096);
  OutputBuf := AllocMem(GetBase64EncodeSize(4096));
  try
    while True do
    begin
      InputBufSize := InputStream.Read(InputBuf^, 4096);
      if InputBufSize <= 0 then
        Break;

      Base64Encode(InputBuf, InputBufSize, OutputBuf, OutputBufSize);
      OutputStream.Write(OutputBuf^, OutputBufSize);
    end;
  finally
    FreeMem(OutputBuf);
    FreeMem(InputBuf);
  end;
end;

procedure Base64DecodeStream(const InputStream: TStream; const OutputStream: TStream);
var
  OutputBuf: PAnsiChar;
  OutputBufSize: Integer;
  TmpSS: TStringStream;
  TmpStr, TmpContinue: AnsiString;
  SurplusLen: Integer;
begin
  TmpStr := '';
  TmpContinue := '';
  TmpSS := TStringStream.Create;
  OutputBuf := AllocMem(GetBase64DecodeSize(4096));
  try
    while True do
    begin
      TmpSS.Clear;

      if InputStream.Size <= InputStream.Position then
        Break;

      if (InputStream.Size-InputStream.Position) < 4096 then
        TmpSS.CopyFrom(InputStream, InputStream.Size-InputStream.Position)
      else
        TmpSS.CopyFrom(InputStream, 4096);

      TmpStr := TmpContinue + TrimIllegalChar(AnsiString(TmpSS.DataString));
      TmpContinue := '';
      if TmpStr = '' then
        Continue;

      if (Length(TmpStr) mod 4) > 0 then
      begin
        SurplusLen := Length(TmpStr) mod 4;
        TmpContinue := Copy(TmpStr, Length(TmpStr) + 1 - SurplusLen, SurplusLen);
        Delete(TmpStr, Length(TmpStr) + 1 - SurplusLen, SurplusLen);
      end;

      Base64Decode(@TmpStr[1], Length(TmpStr), OutputBuf, OutputBufSize);
      OutputStream.Write(OutputBuf^, OutputBufSize);
    end;
  finally
    FreeMem(OutputBuf);
    TmpSS.Free;
  end;
end;

end.
