﻿unit XMLTBase;

interface

uses
  Classes, Dialogs;

type
  TXmlEncodingType = (xetUnknow, xetUTF_8, xetUTF_16,
    xetBIG5, xetGB2312, xetS_JIS, xetEUC_KR, xetIS0_8859_1);

function XmlT_StrToBool(const Value: string): Boolean;
function XmlT_BoolToStr(const Value: Boolean): string;

function XmlT_StrToDateTime(const Value: string): TDateTime;
function XmlT_StrToDate(const Value: string): TDateTime;
function XmlT_StrToTime(const Value: string): TDateTime;
function XmlT_DateTimeToStr(const Value: TDateTime): string;
function XmlT_DateToStr(const Value: TDateTime): string;
function XmlT_TimeToStr(const Value: TDateTime): string;

procedure XmlT_Base64EncodeStream(const InputStream: TStream; const OutputStream: TStream);
procedure XmlT_Base64DecodeStream(const InputStream: TStream; const OutputStream: TStream);

const
  XT_XML_EncodingStr: array[TXmlEncodingType] of string =
  ('Unknow', 'UTF-8', 'UTF-16', 'BIG5', 'GB2312', 'S-JIS', 'EUC-KR', 'IS0-8859-1');
  XT_True_Text: array[0..14] of string =
  ('true', 'T', 'Yes', 'Y', 'Just', 'Right', 'Correct', 'OK', '1', 'On', 'Agree', '是', '對', '好', '正確');
  XT_False_Text: array[0..17] of string =
  ('false', 'F', 'No', 'N', 'Unjust', 'Wrong', 'Incorrect', 'Cancel', '0', 'Off', 'Disagree', '否', '不對',
    '錯', '不', '不好', '不正確', '錯誤');

implementation

uses
  SysUtils, Windows, XSBuiltIns, DateUtils, Base64;

{ XSBuiltIns in $(BDS)\source\Win32\soap\XSBuiltIns.pas }

procedure XmlT_Base64EncodeStream(const InputStream: TStream; const OutputStream: TStream);
begin
  Base64EncodeStream(InputStream, OutputStream);
end;

procedure XmlT_Base64DecodeStream(const InputStream: TStream; const OutputStream: TStream);
begin
  Base64DecodeStream(InputStream, OutputStream);
end;

function NumberToken(var SrcStr: string; IncludeTokenText: Boolean = False): Integer;
var
  i: Integer;
begin
  Result := 0;

  if SrcStr = '' then
    Exit;

  while Length(SrcStr) > 0 do
  begin
    if not CharInSet(SrcStr[1], ['0'..'9']) then
      Delete(SrcStr, 1, 1)
    else
      Break;
  end;

  for i := 1 to Length(SrcStr) do
    if not CharInSet(SrcStr[i], ['0'..'9']) then
    begin
      Result := StrToIntDef(Copy(SrcStr, 1, i - 1), 0);
      Delete(SrcStr, 1, i - 1);
      Exit;
    end;

  Result := StrToIntDef(SrcStr, 0);
  SrcStr := '';
end;

function GetTimeZoneBias: Integer;
var
  TimeZoneInfo: TTimeZoneInformation;
begin
  case GetTimeZoneInformation(TimeZoneInfo) of
    TIME_ZONE_ID_STANDARD: Result := TimeZoneInfo.Bias + TimeZoneInfo.StandardBias;
    TIME_ZONE_ID_DAYLIGHT: Result := TimeZoneInfo.Bias + TimeZoneInfo.DaylightBias;
    TIME_ZONE_ID_INVALID: Result := 0;
  else
    Result := TimeZoneInfo.Bias;
  end;
end;

const
  BoolHourMarker: array [Boolean] of string =
      (XMLHourOffsetMinusMarker, XMLHourOffsetPlusMarker);

function FixDateStr(Str: string): string;
var
  Y, M, D: Word;
begin
  Str := UpperCase(Trim(Str));
  Str := StringReplace(Str, '/', XMLDateSeparator, [rfReplaceAll]);
  Str := StringReplace(Str, '.', XMLDateSeparator, [rfReplaceAll]);
  Str := StringReplace(Str, ',', XMLDateSeparator, [rfReplaceAll]);
  Str := StringReplace(Str, ' ', XMLDateSeparator, [rfReplaceAll]);

  Y := NumberToken(Str);
  M := NumberToken(Str);
  D := NumberToken(Str);
  Result := Format('%.04d%s%.02d%s%.02d%s',
    [Y, XMLDateSeparator, M, XMLDateSeparator, D, Str]);
end;

function FixTimeStr(Str: string): string;
var
  H, N, S, MS: Word;
  TimeZoneBias: Integer;
begin
  Str := UpperCase(Trim(Str));
  Str := StringReplace(Str, '-', XMLTimeSeparator, [rfReplaceAll]);
  Str := StringReplace(Str, '.', XMLTimeSeparator, [rfReplaceAll]);
  Str := StringReplace(Str, ' ', XMLTimeSeparator, [rfReplaceAll]);

  H := NumberToken(Str);
  N := NumberToken(Str);
  S := NumberToken(Str);
  if (Str <> '') and CharInSet(Str[1], ['.'..':']) then
    MS := NumberToken(Str)
  else
    MS := 0;
  if (Str = '') or not CharInSet(Str[1], ['Z', '+', '-'])  then
  begin
    TimeZoneBias := GetTimeZoneBias;
    Str := Format('%s%.02d%s%.02d', [BoolHourMarker[TimeZoneBias < 0],
        (TimeZoneBias div 60)*-1, XMLTimeSeparator, (TimeZoneBias mod 60)*-1]);
  end;

  Result := Format('%.02d%s%.02d%s%.02d%s%.03d%s',
    [H, XMLTimeSeparator, N, XMLTimeSeparator, S, SoapDecimalSeparator, MS, Str]);
end;

function LocaleDateTimeToGMTDateTime(const Value: TDateTime): TDateTime;
begin
  Result := Value + (GetTimeZoneBias / MinsPerDay);
end;

function GMTDateTimeToLocaleDateTime(const Value: TDateTime): TDateTime;
begin
  Result := Value - (GetTimeZoneBias / MinsPerDay);
end;

function XmlT_StrToBool(const Value: string): Boolean;
var
  TestStr: string;
  i: Integer;
begin
  Result := False;
  
  TestStr := UpperCase(Trim(Value));

  for i := Low(XT_True_Text) to High(XT_True_Text) do
    if UpperCase(XT_True_Text[i]) = TestStr then
    begin
      Result := True;
      Break;
    end;
end;

function XmlT_BoolToStr(const Value: Boolean): string;
begin
  if Value then
    Result := XT_True_Text[0]
  else
    Result := XT_False_Text[0];
end;

function XmlT_StrToDateTime(const Value: string): TDateTime;
var
  TestStr, DateStr, TimeStr: string;
begin
  Result := 0;

  TestStr := Trim(Value);

  if TestStr = '' then
    Exit;

  if Pos(' ', TestStr) > 0 then
    TestStr[Pos(' ', TestStr)] := SoapTimePrefix[1];

  if Pos(SoapTimePrefix, TestStr) = 0 then
  begin
    if Pos(XMLTimeSeparator, TestStr) > 0 then
      Result := XmlT_StrToTime(TestStr)
    else
      Result := XmlT_StrToDate(TestStr);
  end
  else
  begin
    try
      DateStr := Copy(TestStr, 1, Pos(SoapTimePrefix, TestStr) - 1);
      TimeStr := Copy(TestStr, Pos(SoapTimePrefix, TestStr) + 1, MaxInt);

      DateStr := FixDateStr(DateStr);
      TimeStr := FixTimeStr(TimeStr);

      Result := XmlT_StrToDate(DateStr) + XmlT_StrToTime(TimeStr);
    except
      Result := 0;
    end;
  end;
end;

function XmlT_StrToDate(const Value: string): TDateTime;
var
  TestStr: string;
  XSDate: TXSDate;
begin
  Result := 0;

  try
    TestStr := Trim(Value);

    if TestStr = '' then
      Exit;

    TestStr := FixDateStr(TestStr);

    XSDate := TXSDate.Create;
    try
      XSDate.XSToNative(TestStr);
      Result := EncodeDate(XSDate.Year, XSDate.Month, XSDate.Day);
    finally
      XSDate.Free;
    end;
  except
    Result := 0;
  end;
end;

function XmlT_StrToTime(const Value: string): TDateTime;
var
  TestStr: string;
  XSTime: TXSTime;
begin
  Result := 0;

  try
    TestStr := Trim(Value);

    if Pos('T', TestStr) > 0 then
      Delete(TestStr, 1, Pos('T', TestStr));

    if TestStr = '' then
      Exit;

    TestStr := FixTimeStr(TestStr);

    XSTime := TXSTime.Create;
    try
      XSTime.UseZeroMilliseconds := False;
      XSTime.XSToNative(TestStr);

      Result := GMTDateTimeToLocaleDateTime(XSTime.AsTime);

      Result := IncHour(Result, -XSTime.HourOffset);
      Result := IncMinute(Result, -XSTime.MinuteOffset);
    finally
      XSTime.Free;
    end;
  except
    on E: Exception do
      Result := 0;
  end;
end;

function XmlT_DateTimeToStr(const Value: TDateTime): string;
  function IsOnlyDate: Boolean;
  var
    D1, D2, D3: Word;
  begin
    DecodeDate(Value, D1, D2, D3);
    Result := Value = EncodeDate(D1, D2, D3);
  end;
  function IsOnlyTime: Boolean;
  var
    D1, D2, D3: Word;
  begin
    DecodeDate(Value, D1, D2, D3);
    Result := (D1 = 1899) and (D2 = 12) and (D3 = 30);
  end;
begin
  Result := '';

  if Value = 0 then
    Result := ''
  else
  begin
    if IsOnlyDate then
      Result := XmlT_DateToStr(Value)
    else if IsOnlyTime then
      Result := XmlT_TimeToStr(Value)
    else
      Result := XmlT_DateToStr(Value) + SoapTimePrefix + XmlT_TimeToStr(Value);
  end;
end;

function XmlT_DateToStr(const Value: TDateTime): string;
var
  XSDate: TXSDate;
begin
  Result := '';

  if Value = 0 then
    Result := ''
  else
  begin
    XSDate := TXSDate.Create;
    try
      XSDate.AsDate := Value;
      Result := XSDate.NativeToXS;
    finally
      XSDate.Free;
    end;
  end;
end;

function XmlT_TimeToStr(const Value: TDateTime): string;
var
  XSTime: TXSTime;
  TimeZoneBias: Integer;
begin
  Result := '';

  if Value = 0 then
    Result := ''
  else
  begin
    XSTime := TXSTime.Create;
    try
      XSTime.UseZeroMilliseconds := False;
      XSTime.AsTime := Value;
      TimeZoneBias := GetTimeZoneBias;
      Result := Format('%.02d%s%.02d%s%.02d%s%.03d%s%.02d%s%.02d',
          [XSTime.Hour, XMLTimeSeparator, XSTime.Minute,
          XMLTimeSeparator, XSTime.Second,
          SoapDecimalSeparator, XSTime.Millisecond, BoolHourMarker[TimeZoneBias < 0],
          Abs(TimeZoneBias div 60), XMLTimeSeparator, Abs(TimeZoneBias mod 60)])
    finally
      XSTime.Free;
    end;
  end;
end;

end.
