unit XMLTBase;

interface

uses
  Classes, Dialogs;

type
  TXmlEncodingType = (xetUnknow, xetUTF_8, xetUTF_16,
    xetBIG5, xetGB2312, xetS_JIS, xetEUC_KR, xetIS0_8859_1);

function XmlT_StrToBool(const Value: WideString): Boolean;
function XmlT_BoolToStr(const Value: Boolean): WideString;

function XmlT_StrToDateTime(const Value: WideString): TDateTime;
function XmlT_StrToDate(const Value: WideString): TDateTime;
function XmlT_StrToTime(const Value: WideString): TDateTime;
function XmlT_DateTimeToStr(const Value: TDateTime): WideString;
function XmlT_DateToStr(const Value: TDateTime): WideString;
function XmlT_TimeToStr(const Value: TDateTime): WideString;

function XmlT_Base64ToStr(Base64: string): WideString;
function XmlT_StrToBase64(Str: WideString): string;

procedure XmlT_MimeEncodeStream(const InputStream: TStream; const OutputStream: TStream);
procedure XmlT_MimeDecodeStream(const InputStream: TStream; const OutputStream: TStream);

const
  XT_XML_EncodingStr: array[TXmlEncodingType] of WideString =
  ('Unknow', 'UTF-8', 'UTF-16', 'BIG5', 'GB2312', 'S-JIS', 'EUC-KR', 'IS0-8859-1');
  XT_True_Text: array[0..14] of WideString =
  ('true', 'T', 'Yes', 'Y', 'Just', 'Right', 'Correct', 'OK', '1', 'On', 'Agree', 'O', '', 'n', 'T');
  XT_False_Text: array[0..17] of WideString =
  ('false', 'F', 'No', 'N', 'Unjust', 'Wrong', 'Incorrect', 'Cancel', '0', 'Off', 'Disagree', '_', '',
    '', '', 'n', 'T', '~');

implementation

uses
  SysUtils, Windows, XSBuiltIns, DateUtils, Base64;

procedure XmlT_MimeEncodeStream(const InputStream: TStream; const OutputStream: TStream);
begin
  MimeEncodeStreamNoCRLF(InputStream, OutputStream);
end;

procedure XmlT_MimeDecodeStream(const InputStream: TStream; const OutputStream: TStream);
begin
  MimeDecodeStream(InputStream, OutputStream);
end;

function NumberToken(var SrcStr: WideString; IncludeTokenText: Boolean = False): Integer;
var
  i: Integer;
begin
  Result := 0;

  if SrcStr = '' then
    Exit;

  while Length(SrcStr) > 0 do
  begin
    if not (SrcStr[1] in [WideChar('0')..WideChar('9')]) then
      Delete(SrcStr, 1, 1)
    else
      Break;
  end;

  for i := 1 to Length(SrcStr) do
    if not (SrcStr[i] in [WideChar('0')..WideChar('9')]) then
    begin
      Result := StrToIntDef(Copy(SrcStr, 1, i - 1), 0);
      Delete(SrcStr, 1, i - 1);
      Exit;
    end;

  Result := StrToIntDef(SrcStr, 0);
  SrcStr := '';
end;

function GetTimeZoneBias: Integer;
var
  TimeZoneInfo: TTimeZoneInformation;
begin
  case GetTimeZoneInformation(TimeZoneInfo) of
    TIME_ZONE_ID_STANDARD: Result := TimeZoneInfo.Bias + TimeZoneInfo.StandardBias;
    TIME_ZONE_ID_DAYLIGHT: Result := TimeZoneInfo.Bias + TimeZoneInfo.DaylightBias;
    TIME_ZONE_ID_INVALID: Result := 0;
  else
    Result := TimeZoneInfo.Bias;
  end;
end;

const
  BoolHourMarker: array [Boolean] of string =
      (XMLHourOffsetMinusMarker, XMLHourOffsetPlusMarker);

function FixDateStr(Str: WideString): WideString;
var
  Y, M, D: Word;
begin
  Str := UpperCase(Trim(Str));
  Str := StringReplace(Str, '/', XMLDateSeparator, [rfReplaceAll]);
  Str := StringReplace(Str, '.', XMLDateSeparator, [rfReplaceAll]);
  Str := StringReplace(Str, ',', XMLDateSeparator, [rfReplaceAll]);
  Str := StringReplace(Str, ' ', XMLDateSeparator, [rfReplaceAll]);

  Y := NumberToken(Str);
  M := NumberToken(Str);
  D := NumberToken(Str);
  Result := Format('%.04d%s%.02d%s%.02d%s',
    [Y, XMLDateSeparator, M, XMLDateSeparator, D, Str]);
end;

function FixTimeStr(Str: WideString): WideString;
var
  H, N, S, MS: Word;
  TimeZoneBias: Integer;
begin
  Str := UpperCase(Trim(Str));
  Str := StringReplace(Str, '-', XMLTimeSeparator, [rfReplaceAll]);
  Str := StringReplace(Str, '.', XMLTimeSeparator, [rfReplaceAll]);
  Str := StringReplace(Str, ' ', XMLTimeSeparator, [rfReplaceAll]);

  H := NumberToken(Str);
  N := NumberToken(Str);
  S := NumberToken(Str);
  if (Str <> '') and (Str[1] in [WideChar('.'), WideChar(':')]) then
    MS := NumberToken(Str)
  else
    MS := 0;
  if (Str = '') or not (Str[1] in [WideChar('Z'), WideChar('+'), WideChar('-')])  then
  begin
    TimeZoneBias := GetTimeZoneBias;
    Str := Format('%s%.02d%s%.02d', [BoolHourMarker[TimeZoneBias < 0],
        (TimeZoneBias div 60)*-1, XMLTimeSeparator, (TimeZoneBias mod 60)*-1]);
  end;

  Result := Format('%.02d%s%.02d%s%.02d%s%.03d%s',
    [H, XMLTimeSeparator, N, XMLTimeSeparator, S, SoapDecimalSeparator, MS, Str]);
end;

function LocaleDateTimeToGMTDateTime(const Value: TDateTime): TDateTime;
begin
  Result := Value + (GetTimeZoneBias / MinsPerDay);
end;

function GMTDateTimeToLocaleDateTime(const Value: TDateTime): TDateTime;
begin
  Result := Value - (GetTimeZoneBias / MinsPerDay);
end;

function XmlT_StrToBool(const Value: WideString): Boolean;
var
  TestStr: WideString;
  i: Integer;
begin
  Result := False;
  
  TestStr := UpperCase(Trim(Value));

  for i := Low(XT_True_Text) to High(XT_True_Text) do
    if UpperCase(XT_True_Text[i]) = TestStr then
    begin
      Result := True;
      Break;
    end;
end;

function XmlT_BoolToStr(const Value: Boolean): WideString;
begin
  if Value then
    Result := XT_True_Text[0]
  else
    Result := XT_False_Text[0];
end;

function XmlT_StrToDateTime(const Value: WideString): TDateTime;
var
  TestStr, DateStr, TimeStr: WideString;
begin
  Result := 0;

  TestStr := Trim(Value);

  if TestStr = '' then
    Exit;

  if Pos(' ', TestStr) > 0 then
    TestStr[Pos(' ', TestStr)] := SoapTimePrefix;

  if Pos(SoapTimePrefix, TestStr) = 0 then
  begin
    if Pos(XMLTimeSeparator, TestStr) > 0 then
      Result := XmlT_StrToTime(TestStr)
    else
      Result := XmlT_StrToDate(TestStr);
  end
  else
  begin
    try
      DateStr := Copy(TestStr, 1, Pos(SoapTimePrefix, TestStr) - 1);
      TimeStr := Copy(TestStr, Pos(SoapTimePrefix, TestStr) + 1, MaxInt);

      DateStr := FixDateStr(DateStr);
      TimeStr := FixTimeStr(TimeStr);

      Result := XmlT_StrToDate(DateStr) + XmlT_StrToTime(TimeStr);
    except
      Result := 0;
    end;
  end;
end;

function XmlT_StrToDate(const Value: WideString): TDateTime;
var
  TestStr: WideString;
  XSDate: TXSDate;
begin
  Result := 0;

  try
    TestStr := Trim(Value);

    if TestStr = '' then
      Exit;

    TestStr := FixDateStr(TestStr);

    XSDate := TXSDate.Create;
    try
      XSDate.XSToNative(TestStr);
      Result := EncodeDate(XSDate.Year, XSDate.Month, XSDate.Day);
    finally
      XSDate.Free;
    end;
  except
    Result := 0;
  end;
end;

function XmlT_StrToTime(const Value: WideString): TDateTime;
var
  TestStr: WideString;
  XSTime: TXSTime;
begin
  Result := 0;

  try
    TestStr := Trim(Value);

    if Pos('T', TestStr) > 0 then
      Delete(TestStr, 1, Pos('T', TestStr));

    if TestStr = '' then
      Exit;

    TestStr := FixTimeStr(TestStr);

    XSTime := TXSTime.Create;
    try
      XSTime.UseZeroMilliseconds := False;
      XSTime.XSToNative(TestStr);

      Result := GMTDateTimeToLocaleDateTime(XSTime.AsTime);

      Result := IncHour(Result, -XSTime.HourOffset);
      Result := IncMinute(Result, -XSTime.MinuteOffset);
    finally
      XSTime.Free;
    end;
  except
    on E: Exception do
      Result := 0;
  end;
end;

function XmlT_DateTimeToStr(const Value: TDateTime): WideString;
  function IsOnlyDate: Boolean;
  var
    D1, D2, D3: Word;
  begin
    DecodeDate(Value, D1, D2, D3);
    Result := Value = EncodeDate(D1, D2, D3);
  end;
  function IsOnlyTime: Boolean;
  var
    D1, D2, D3: Word;
  begin
    DecodeDate(Value, D1, D2, D3);
    Result := (D1 = 1899) and (D2 = 12) and (D3 = 30);
  end;
begin
  Result := '';

  if Value = 0 then
    Result := ''
  else
  begin
    if IsOnlyDate then
      Result := XmlT_DateToStr(Value)
    else if IsOnlyTime then
      Result := XmlT_TimeToStr(Value)
    else
      Result := XmlT_DateToStr(Value) + SoapTimePrefix + XmlT_TimeToStr(Value);
  end;
end;

function XmlT_DateToStr(const Value: TDateTime): WideString;
var
  XSDate: TXSDate;
begin
  Result := '';

  if Value = 0 then
    Result := ''
  else
  begin
    XSDate := TXSDate.Create;
    try
      XSDate.AsDate := Value;
      Result := XSDate.NativeToXS;
    finally
      XSDate.Free;
    end;
  end;
end;

function XmlT_TimeToStr(const Value: TDateTime): WideString;
var
  XSTime: TXSTime;
  TimeZoneBias: Integer;
begin
  Result := '';

  if Value = 0 then
    Result := ''
  else
  begin
    XSTime := TXSTime.Create;
    try
      XSTime.UseZeroMilliseconds := False;
      XSTime.AsTime := Value;
      TimeZoneBias := GetTimeZoneBias;
      Result := Format('%.02d%s%.02d%s%.02d%s%.03d%s%.02d%s%.02d',
          [XSTime.Hour, XMLTimeSeparator, XSTime.Minute,
          XMLTimeSeparator, XSTime.Second,
          SoapDecimalSeparator, XSTime.Millisecond, BoolHourMarker[TimeZoneBias < 0],
          Abs(TimeZoneBias div 60), XMLTimeSeparator, Abs(TimeZoneBias mod 60)])
    finally
      XSTime.Free;
    end;
  end;
end;

function XmlT_Base64ToStr(Base64: string): WideString;
var
  S1: TMemoryStream;
  S2: TStringStream;
begin
  S2 := TStringStream.Create(Trim(Base64));
  try
    S1 := TMemoryStream.Create;
    try
      S2.Position := 0;
      XmlT_MimeDecodeStream(S2, S1);
      SetLength(Result, S1.Size + 2);
      S1.Read(Result, S1.Size);
    finally
      S1.Free;
    end;
  finally
    S2.Free;
  end;
end;

function XmlT_StrToBase64(Str: WideString): string;
var
  S1: TMemoryStream;
  S2: TStringStream;
begin
  S1 := TMemoryStream.Create;
  try
    S1.Write(Str, Length(Str));
    S2 := TStringStream.Create('');
    try
      S1.Position := 0;
      XmlT_MimeEncodeStream(S1, S2);
      Result := S2.DataString;
    finally
      S2.Free;
    end;
  finally
    S1.Free;
  end;
end;

end.
