unit XmlIni;

interface

{
  Design By Howard Chen.
  EMail: powerhowardchen@gmail.com
  MSN: powerhowardchen@hotmail.com

  XmlIni version 5.7.

  This header may not be removed.
}

{
  AES Functions.
  Copyright (c) 1998-2001
  EldoS, Alexander Ionov
}

// Compiler with RSA Encrypt function
{$DEFINE RSA_ENCRYPT_ENABLED}

// Compiler with AES Encrypt function
{$DEFINE AES_ENCRYPT_ENABLED}

uses
  Windows, Classes, XMLIntf, Dialogs, XMLTBase
  {$IFDEF RSA_ENCRYPT_ENABLED}
  , RSAUtil
  {$ENDIF}
  ;

{$IFDEF RSA_ENCRYPT_ENABLED}
  {$DEFINE AES_ENCRYPT_ENABLED}
{$ENDIF}

type
  TXmlIni = class;
  TXmlIniSection = class;
  TXmlIniKey = class;

  TXmlIniDataType = (xidtUnknow,
      xidtText, xidtInteger, xidtFloat, xidtBool, xidtDateTime, xidtDate, xidtTime, xidtBinary);

  TXmlIniMediaType = (ximtString, ximtBase64String, ximtFile, ximtRegistry, ximtIni, ximtShareMem);

  TXmlIniEncryptMode = (xiemAuto, xiemPlain
{$IFDEF AES_ENCRYPT_ENABLED}
  , xiemAES, xiemAES_E1, xiemAES_E2
{$ENDIF}
{$IFDEF RSA_ENCRYPT_ENABLED}
  , xiemRSA, xiemRSA_E1, xiemRSA_E2, xiemRSA_E3
{$ENDIF}
  );

  TXmlIniEventMemoryStream = class(TMemoryStream)
  private
    FOwner: TXmlIniKey;
  public
    constructor Create(AOwner: TXmlIniKey);
    function GetSize: Int64; override;
    procedure SetSize(NewSize: Longint); override;
    function Read(var Buffer; Count: Longint): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;
  end;

  TXmlIniKey = class(TObject)
  private
    FOwner: TXmlIniSection;
    FName: WideString;
    FText: WideString;
    FStream: TXmlIniEventMemoryStream;
    FDataType: TXmlIniDataType;
    FRaiseException: Boolean;
  protected
    procedure BinaryToText; virtual;
    procedure TextToBinary; virtual;

    function DetectDataType(Data: WideString): TXmlIniDataType;

    procedure SetName(Value: WideString); virtual;
    function GetText: WideString; virtual;
    procedure SetText(const Value: WideString); virtual;
    function GetAsStream: TStream; virtual;
    procedure SetAsStream(const Value: TStream); virtual;
    function GetAsInteger: Integer; virtual;
    procedure SetAsInteger(const Value: Integer); virtual;
    function GetAsFloat: Double; virtual;
    procedure SetAsFloat(const Value: Double); virtual;
    function GetAsBool: Boolean; virtual;
    procedure SetAsBool(const Value: Boolean); virtual;
    function GetAsDateTime: TDateTime; virtual;
    procedure SetAsDateTime(const Value: TDateTime); virtual;
    function GetAsDate: TDateTime; virtual;
    procedure SetAsDate(const Value: TDateTime); virtual;
    function GetAsTime: TDateTime; virtual;
    procedure SetAsTime(const Value: TDateTime); virtual;
    function GetDataType: TXmlIniDataType; virtual;
    procedure SetDataType(const Value: TXmlIniDataType); virtual;
    function GetIndex: Integer; virtual;
  public
    constructor Create(AName: WideString = ''; AText: WideString = ''; AOwner: TXmlIniSection = nil); virtual;
    destructor Destroy; override;

    procedure Clear;

    procedure ValueLoadFromFile(FileName: WideString; Binary: Boolean = True); virtual;
    procedure ValueSaveToFile(FileName: WideString); virtual;

    property Name: WideString read FName write SetName;
    property Text: WideString read GetText write SetText;

    property AsStream: TStream read GetAsStream write SetAsStream;
    property AsInteger: Integer read GetAsInteger write SetAsInteger;
    property AsFloat: Double read GetAsFloat write SetAsFloat;
    property AsBool: Boolean read GetAsBool write SetAsBool;
    property AsDateTime: TDateTime read GetAsDateTime write SetAsDateTime;
    property AsDate: TDateTime read GetAsDate write SetAsDate;
    property AsTime: TDateTime read GetAsTime write SetAsTime;

    property DataType: TXmlIniDataType read GetDataType write SetDataType;

    property Index: Integer read GetIndex;

    property RaiseException: Boolean read FRaiseException write FRaiseException;
  end;

  TXmlIniSection = class(TObject)
  private
    FName: WideString;
    FOwner: TXmlIni;

    FList: TList;
    FSortedList: TList;

    FXmlEncoding: TXmlEncodingType;
    FCaseSensitive: Boolean;
    FCompressed: Boolean;
    FRaiseException: Boolean;
  protected
    procedure RegisterKey(Key: TXmlIniKey); virtual;
    procedure UnregisterKey(Key: TXmlIniKey); virtual;

    function InternalFindCaseSensitive(const S: WideString; var Index: Integer): Boolean; virtual;
    function InternalFindCaseInsensitive(const S: WideString; var Index: Integer): Boolean; virtual;
    function InternalFind(const KeyName: WideString; var Index: Integer): Boolean; virtual;

    function InternalAdd(KeyName: WideString): TXmlIniKey; virtual;

    procedure SetCaseSensitive(Value: Boolean); virtual;
    procedure SetName(Value: WideString); virtual;
    function GetKeyByIndex(Index: Integer): TXmlIniKey; virtual;
    function GetKey(KeyName: WideString): TXmlIniKey; virtual;
    function GetItem(Index: Variant): WideString; virtual;
    procedure SetItem(Index: Variant; const Value: WideString); virtual;
    function GetCount: Integer; virtual;
    function GetIndex: Integer; virtual;
    procedure SetRaiseException(const Value: Boolean); virtual;
  public
    constructor Create(AName: WideString = ''; AOwner: TXmlIni = nil); virtual;
    destructor Destroy; override;

    function Equals(Dest: TXmlIniSection): Boolean; overload;
    procedure CopyFrom(Source: TXmlIniSection); overload;
    procedure Append(Source: TXmlIniSection); overload;

    function Equals(Dest: TStrings): Boolean; overload;
    procedure CopyFrom(Source: TStrings); overload;
    procedure Append(Source: TStrings); overload;

    function Exists(KeyName: WideString): Boolean;

    procedure Clear;
    procedure Collation;

    procedure LoadFromStream(Stream: TStream; SectionName: WideString = ''); virtual;
    procedure SaveToStream(Stream: TStream; SectionName: WideString = ''); virtual;

    procedure LoadFromString(Value: WideString; SectionName: WideString = '');
    function SaveToString(SectionName: WideString = ''): WideString;

    procedure LoadFromBase64String(Value: string; SectionName: WideString = '');
    function SaveToBase64String(SectionName: WideString = ''): string;

    procedure LoadFromFile(FileName: WideString; SectionName: WideString = '');
    procedure SaveToFile(FileName: WideString; SectionName: WideString = '');
    procedure UpdateToFile(FileName: WideString; SectionName: WideString = '');

    procedure LoadFromRegistry(Path: WideString; RootKey: HKEY = HKEY_LOCAL_MACHINE; SectionName: WideString = ''); virtual;
    procedure SaveToRegistry(Path: WideString; RootKey: HKEY = HKEY_LOCAL_MACHINE; SectionName: WideString = '';
      UseOldDataType: Boolean = False); virtual;
    procedure UpdateToRegistry(Path: WideString; RootKey: HKEY = HKEY_LOCAL_MACHINE; SectionName: WideString = '';
      UseOldDataType: Boolean = False); virtual;

    procedure LoadFromIni(FileName: WideString; SectionName: WideString = '');
    procedure SaveToIni(FileName: WideString; SectionName: WideString = '');

    property XmlEncoding: TXmlEncodingType read FXmlEncoding write FXmlEncoding;
    property CaseSensitive: Boolean read FCaseSensitive write SetCaseSensitive;
    property Compressed: Boolean read FCompressed write FCompressed;

    property Name: WideString read FName write SetName;
    property KeyByIndex[Index: Integer]: TXmlIniKey read GetKeyByIndex;
    property Key[KeyName: WideString]: TXmlIniKey read GetKey;
    property Item[Index: Variant]: WideString read GetItem write SetItem; default;

    property Count: Integer read GetCount;

    property Index: Integer read GetIndex;

    property RaiseException: Boolean read FRaiseException write SetRaiseException;
  end;

  TXmlIni = class(TComponent)
  private
    FList: TList;
    FSortedList: TList;

    FXmlEncoding: TXmlEncodingType;

    FGlobalID: WideString;
    FFileName: WideString;
    FAutoSave: Boolean;
    FCaseSensitive: Boolean;
    FCompressed: Boolean;
    FEncryptMode: TXmlIniEncryptMode;
    {$IFDEF AES_ENCRYPT_ENABLED}
    FPassword: WideString;
    {$ENDIF}
    {$IFDEF RSA_ENCRYPT_ENABLED}
    FPrivateKey: TStringList;
    FPublicKey: TStringList;
    FSignPrivateKey: TStringList;
    FVerifySignPublicKey: TStringList;
    {$ENDIF}
    FRaiseException: Boolean;

    FLockCriSec: TRTLCriticalSection;

    procedure ExtractRegPath(var Path: WideString; var RootKey: HKEY); virtual;
  protected
    procedure RegisterSection(Section: TXmlIniSection); virtual;
    procedure UnregisterSection(Section: TXmlIniSection); virtual;

    procedure InternalXmlLoadFromStream(Stream: TStream); virtual;
    procedure InternalXmlSaveToStream(Stream: TStream); virtual;
    procedure InternalLoadFromFile(AFileName: WideString); virtual;
    procedure InternalSaveToFile(AFileName: WideString); virtual;

    function InternalFindCaseSensitive(const S: WideString; var Index: Integer): Boolean; virtual;
    function InternalFindCaseInsensitive(const S: WideString; var Index: Integer): Boolean; virtual;
    function InternalFind(const SectionName: WideString; var Index: Integer): Boolean; virtual;

    function InternalAdd(SectionName: WideString): TXmlIniSection; virtual;

    function GetCount: Integer;
    function GetSectionByIndex(Index: Integer): TXmlIniSection;
    function GetSection(Index: Variant): TXmlIniSection;

    procedure SetGlobalID(const Value: WideString);
    procedure SetFileName(const Value: WideString);
    procedure SetAutoSave(const Value: Boolean);
    procedure SetCaseSensitive(const Value: Boolean);
    procedure SetCompressed(const Value: Boolean);
    procedure SetEncryptMode(const Value: TXmlIniEncryptMode);
    {$IFDEF AES_ENCRYPT_ENABLED}
    procedure SetPassword(const Value: WideString);
    {$ENDIF}
    {$IFDEF RSA_ENCRYPT_ENABLED}
    function GetPrivateKey: TStrings;
    procedure SetPrivateKey(const Value: TStrings);
    function GetPublicKey: TStrings;
    procedure SetPublicKey(const Value: TStrings);
    function GetSignPrivateKey: TStrings;
    procedure SetSignPrivateKey(const Value: TStrings);
    function GetVerifySignPublicKey: TStrings;
    procedure SetVerifySignPublicKey(const Value: TStrings);
    {$ENDIF}
    procedure SetRaiseException(const Value: Boolean);
  public
    constructor Create(AOwner: TComponent); overload; override;
    constructor Create(AFileName: WideString = ''); reintroduce; overload;
    destructor Destroy; override;

    function Equals(Dest: TXmlIni): Boolean; virtual;
    procedure CopyFrom(Source: TXmlIni); virtual;
    procedure Append(Source: TXmlIni); virtual;

    function Add(SectionName: WideString): TXmlIniSection; virtual;
    procedure Delete(SectionName: WideString); virtual;
    function Exists(SectionName: WideString; KeyName: WideString = ''): Boolean; virtual;

    procedure Clear; virtual;
    procedure Collation; virtual;

    procedure LoadFromStream(Stream: TStream); virtual;
    procedure SaveToStream(Stream: TStream); virtual;

    procedure LoadFromString(Value: WideString); virtual;
    function SaveToString: WideString; virtual;

    procedure LoadFromBase64String(Value: string); virtual;
    function SaveToBase64String: string; virtual;

    procedure LoadFromFile(AFileName: WideString = ''); virtual;
    procedure SaveToFile(AFileName: WideString = ''); virtual;
    procedure UpdateToFile(AFileName: WideString = ''); virtual;

    procedure LoadFromRegistry(Path: WideString = ''; RootKey: HKEY = HKEY_LOCAL_MACHINE); virtual;
    procedure SaveToRegistry(Path: WideString = ''; RootKey: HKEY = HKEY_LOCAL_MACHINE); virtual;
    procedure UpdateToRegistry(Path: WideString = ''; RootKey: HKEY = HKEY_LOCAL_MACHINE); virtual;

    procedure LoadFromIni(AIniFileName: WideString = ''); virtual;
    procedure SaveToIni(AIniFileName: WideString = ''); virtual;
    procedure UpdateToIni(AIniFileName: WideString = ''); virtual;

    procedure LoadFromShareMem(ShareMemID: string = ''); virtual;
    procedure SaveToShareMem(ShareMemID: string = ''); virtual;
    procedure UpdateToShareMem(ShareMemID: string = ''); virtual;

    procedure LoadFromMedia(MediaType: TXmlIniMediaType; MediaID: string = ''); virtual;
    function SaveToMedia(MediaType: TXmlIniMediaType; MediaID: string = ''): WideString; virtual;
    function UpdateToMedia(MediaType: TXmlIniMediaType; MediaID: string = ''): WideString; virtual;

    {$IFDEF RSA_ENCRYPT_ENABLED}
    procedure GenerateNewKey; virtual;
    procedure GenerateNewEncryptKey; virtual;
    procedure GenerateNewSignKey; virtual;
    {$ENDIF}

    procedure Lock; virtual;
    procedure Unlock; virtual;

    // TIniFile Functions Begin
    function SectionExists(const Section: WideString): Boolean;

    function ReadString(const Section, Ident: WideString; Default: WideString = ''): WideString; virtual;
    procedure WriteString(const Section, Ident: WideString; Value: WideString); virtual;
    function ReadInteger(const Section, Ident: WideString; Default: Longint = 0): Longint; virtual;
    procedure WriteInteger(const Section, Ident: WideString; Value: Longint); virtual;
    function ReadFloat(const Section, Ident: WideString; Default: Double = 0): Double; virtual;
    procedure WriteFloat(const Section, Ident: WideString; Value: Double); virtual;
    function ReadBool(const Section, Ident: WideString; Default: Boolean = False): Boolean; virtual;
    procedure WriteBool(const Section, Ident: WideString; Value: Boolean); virtual;
    function ReadDateTime(const Section, Ident: WideString; Default: TDateTime = 0): TDateTime; virtual;
    procedure WriteDateTime(const Section, Ident: WideString; Value: TDateTime); virtual;
    function ReadDate(const Section, Ident: WideString; Default: TDateTime = 0): TDateTime; virtual;
    procedure WriteDate(const Section, Ident: WideString; Value: TDateTime); virtual;
    function ReadTime(const Section, Ident: WideString; Default: TDateTime = 0): TDateTime; virtual;
    procedure WriteTime(const Section, Ident: WideString; Value: TDateTime); virtual;
    function ReadBinaryStream(const Section, Ident: WideString; Value: TStream): Integer; virtual;
    procedure WriteBinaryStream(const Section, Ident: WideString; Value: TStream); virtual;

    procedure ReadSection(const Section: WideString; Strings: TStrings); virtual;
    procedure ReadSections(Strings: TStrings); virtual;
    procedure ReadSectionValues(const Section: WideString; Strings: TStrings); virtual;
    procedure EraseSection(const Section: WideString); virtual;

    procedure DeleteKey(const Section, Ident: WideString); virtual;
    function ValueExists(const Section, Ident: WideString): Boolean; virtual;
    // TIniFile Functions End

    property SectionByIndex[Index: Integer]: TXmlIniSection read GetSectionByIndex;
    property Section[Index: Variant]: TXmlIniSection read GetSection; default;
    property Count: Integer read GetCount;
  published
    property GlobalID: WideString read FGlobalID write SetGlobalID;
    property FileName: WideString read FFileName write SetFileName;
    property AutoSave: Boolean read FAutoSave write SetAutoSave default False;

    property XmlEncoding: TXmlEncodingType read FXmlEncoding write FXmlEncoding default xetUTF_16;
    property CaseSensitive: Boolean read FCaseSensitive write SetCaseSensitive default False;
    property Compressed: Boolean read FCompressed write SetCompressed default False;
    property EncryptMode: TXmlIniEncryptMode read FEncryptMode write SetEncryptMode default xiemAuto;
    {$IFDEF AES_ENCRYPT_ENABLED}
    property Password: WideString read FPassword write SetPassword;
    {$ENDIF}
    {$IFDEF RSA_ENCRYPT_ENABLED}
    property PrivateKey: TStrings read GetPrivateKey write SetPrivateKey;
    property PublicKey: TStrings read GetPublicKey write SetPublicKey;
    property SignPrivateKey: TStrings read GetSignPrivateKey write SetSignPrivateKey;
    property VerifySignPublicKey: TStrings read GetVerifySignPublicKey write SetVerifySignPublicKey;
    {$ENDIF}

    property RaiseException: Boolean read FRaiseException write SetRaiseException default False;
  end;

implementation

uses
  xmldom, msxmldom, XMLDoc, XSBuiltIns, ActiveX, Forms, zlib, SysUtils, DateUtils, StrUtils, Registry, IniFiles,
  {$IFDEF AES_ENCRYPT_ENABLED}ElAES, {$ENDIF} Variants;

const
  XI_XML_StandAlone = 'yes';
  XI_XML_Version = '1.0';

  XI_Root_NodeName = 'xmlini';
  XI_Section_NodeName = 'section';
  XI_Key_NodeName = 'key';
  XI_Name_Attribute = 'name';

  XI_Base64Header = '?BASE64?';

  XI_LockMutexName = 'XMLINI_MUTEX_LOCK_';
  XI_ShareName = 'XMLINI_SHARED_MEM_';

const
  BufSize = 1024;

  HeaderCompress = 'ZL';

{$IFDEF RSA_ENCRYPT_ENABLED}
  FakePasswordLength = 16;
  HeaderRSACrypto = 'PC';
  HeaderRSACrypto_LV1 = 'P1';
  HeaderRSACrypto_LV2 = 'P2';
  HeaderRSACrypto_LV3 = 'P3';
{$ENDIF}

{$IFDEF AES_ENCRYPT_ENABLED}
  HeaderAESCrypto = 'EC';
  HeaderAESCrypto_LV1 = 'E1';
  HeaderAESCrypto_LV2 = 'E2';
{$ENDIF}

procedure CompressStream(SourceStream, DestStream: TStream);
var
  CS: TCompressionStream;
  Buf: PChar;
  DataSize, ReadCount: LongInt;
begin
  DataSize := SourceStream.Size;
  DestStream.Write(HeaderCompress, 2);
  DestStream.Write(DataSize, SizeOf(LongInt));

  CS := TCompressionStream.Create(clMax, DestStream);
  try
    Buf := AllocMem(BufSize);
    try
      repeat
        ReadCount := SourceStream.Read(Buf^, BufSize);
        if ReadCount > 0 then
          CS.Write(Buf^, ReadCount);
      until ReadCount = 0;
    finally
      FreeMem(Buf);
    end;
  finally
    CS.Free;
  end;
end;

procedure DecompressStream(SourceStream, DestStream: TStream);
var
  DS: TDecompressionStream;
  Ident: array[0..1] of Char;
  Buf: PChar;
  DataSize, ReadCount: LongInt;
begin
  SourceStream.Read(Ident, 2);
  SourceStream.Read(DataSize, SizeOf(LongInt));
  if (Ident = HeaderCompress) and (DataSize > 0) then
  begin
    DestStream.Size := 0;
    DS := TDecompressionStream.Create(SourceStream);
    try
      Buf := AllocMem(BufSize);
      try
        while DataSize > 0 do
        begin
          ReadCount := DataSize;
          if ReadCount > BufSize then
            ReadCount := BufSize;
          DS.Read(Buf^, ReadCount);
          DestStream.Write(Buf^, ReadCount);
          Dec(DataSize, ReadCount);
        end;
      finally
        FreeMem(Buf);
      end;
    finally
      DS.Free;
    end;
  end;
end;

function RandomStr10Base(StrLen: Integer): string;
var
  Ch, LastCh: Char;
begin
  Result := '';
  LastCh := ' ';
  while Length(Result) < StrLen do
  begin
    Ch := Char(Random($0A)+$30);
    if Ch <> LastCh then
    begin
      Result := Result + Ch;
      LastCh := Ch;
    end;
  end;
end;

function RandomStr256Base(StrLen: Integer): string;
var
  Ch, LastCh: Char;
begin
  Result := '';
  LastCh := ' ';
  while Length(Result) < StrLen do
  begin
    Ch := Char(Random($5C)+$21);
    if Ch <> LastCh then
    begin
      Result := Result + Ch;
      LastCh := Ch;
    end;
  end;
end;

function ExtractFileNameWithoutExt(FileName: string): string;
var
  I: Integer;
begin
  Result := ExtractFileName(FileName);
  I := LastDelimiter('.' + PathDelim + DriveDelim, Result);
  if (I > 0) and (Result[I] = '.') then
    Result := Copy(Result, 1, I - 1)
  else
    Result := Result;
end;

function TextToSingleLine(Str: WideString): WideString;
begin
  Result := Str;
  if (Pos(#13, Result) > 0) or (Pos(#10, Result) > 0) then
  begin
    Result := StringReplace(Result, '&', '&amp;', [rfReplaceAll]);
    Result := StringReplace(Result, #10, '&#xA;', [rfReplaceAll]);
    Result := StringReplace(Result, #13, '&#xD;', [rfReplaceAll]);
  end;
end;

function SingleLineToText(Str: WideString): WideString;
begin
  Result := Str;

  if Pos('&', Result) > 0 then
  begin
    Result := StringReplace(Result, '&#xD;', #13, [rfIgnoreCase, rfReplaceAll]);
    Result := StringReplace(Result, '&#xA;', #10, [rfIgnoreCase, rfReplaceAll]);
    Result := StringReplace(Result, '&amp;', '&', [rfIgnoreCase, rfReplaceAll]);
    Result := StringReplace(Result, '&', '&amp;', [rfReplaceAll]);
  end;
end;

type
  THandledObject = class(TObject)
  protected
    FHandle: THandle;
  public
    destructor Destroy; override;
    property Handle: THandle read FHandle;
  end;

  TMutex = class(THandledObject)
  public
    constructor Create(const Name: string);
    function Get(TimeOut: Integer = -1): Boolean;
    function Release: Boolean;
  end;

{ THandledObject }

destructor THandledObject.Destroy;
begin
  if FHandle <> 0 then
    CloseHandle(FHandle);
end;

{ TMutex }

constructor TMutex.Create(const Name: string);
begin
  FHandle := CreateMutex(nil, False, PChar(Name));
  if FHandle = 0 then
    abort;
end;

function TMutex.Get(TimeOut: Integer = -1): Boolean;
begin
  Result := WaitForSingleObject(FHandle, TimeOut) = WAIT_OBJECT_0;
end;

function TMutex.Release: Boolean;
begin
  Result := ReleaseMutex(FHandle);
end;

{ TXmlIni }

constructor TXmlIni.Create(AFileName: WideString);
begin
  Randomize;

  FList := TList.Create;
  FSortedList := TList.Create;
  FXmlEncoding := xetUTF_16;
  FAutoSave := False;
  FFileName := '';
  FCaseSensitive := False;
  FCompressed := False;
  FEncryptMode := xiemAuto;
  {$IFDEF AES_ENCRYPT_ENABLED}
  FPassword := '';
  {$ENDIF}
  {$IFDEF RSA_ENCRYPT_ENABLED}
  FPrivateKey := TStringList.Create;
  FPublicKey := TStringList.Create;
  FSignPrivateKey := TStringList.Create;
  FVerifySignPublicKey := TStringList.Create;
  {$ENDIF}
  FRaiseException := False;

  try
    LoadFromFile(AFileName);
  except
  end;
  FFileName := AFileName;

  InitializeCriticalSection(FLockCriSec);
end;

constructor TXmlIni.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  CoInitialize(nil);

  Randomize;

  FList := TList.Create;
  FSortedList := TList.Create;
  FXmlEncoding := xetUTF_16;
  FAutoSave := False;
  FFileName := '';
  FCaseSensitive := False;
  FCompressed := False;
  FEncryptMode := xiemAuto;
  {$IFDEF AES_ENCRYPT_ENABLED}
  FPassword := '';
  {$ENDIF}
  {$IFDEF RSA_ENCRYPT_ENABLED}
  FPrivateKey := TStringList.Create;
  FPublicKey := TStringList.Create;
  FSignPrivateKey := TStringList.Create;
  FVerifySignPublicKey := TStringList.Create;
  {$ENDIF}
  FRaiseException := False;

  InitializeCriticalSection(FLockCriSec);
end;

destructor TXmlIni.Destroy;
begin
  DeleteCriticalSection(FLockCriSec);

  if FAutoSave then
    SaveToFile;

  Clear;

  {$IFDEF RSA_ENCRYPT_ENABLED}
  FPrivateKey.Free;
  FPublicKey.Free;
  FSignPrivateKey.Free;
  FVerifySignPublicKey.Free;
  {$ENDIF}

  FSortedList.Free;
  FList.Free;
end;

function TXmlIni.Equals(Dest: TXmlIni): Boolean;
var
  i: Integer;
begin
  Result := False;

  for i := 0 to Count - 1 do
  begin
    if not Dest.Exists(SectionByIndex[i].Name) then
      Exit;
    if not SectionByIndex[i].Equals(Dest[SectionByIndex[i].Name]) then
      Exit;
  end;

  Result := True;
end;

procedure TXmlIni.CopyFrom(Source: TXmlIni);
begin
  Clear;
  Append(Source);
end;

procedure TXmlIni.Append(Source: TXmlIni);
var
  i: Integer;
begin
  Source.Collation;
  for i := 0 to Source.Count - 1 do
    Add(Source.SectionByIndex[i].Name).Append(Source.SectionByIndex[i]);
  Collation;
end;

function TXmlIni.Add(SectionName: WideString): TXmlIniSection;
var
  Index: Integer;
begin
  if InternalFind(SectionName, Index) then
    Result := TXmlIniSection(FSortedList[Index])
  else
    Result := InternalAdd(SectionName);
end;

procedure TXmlIni.Delete(SectionName: WideString);
var
  Index: Integer;
begin
  if InternalFind(SectionName, Index) then
    TXmlIniSection(FSortedList[Index]).Free;
end;

function TXmlIni.Exists(SectionName: WideString; KeyName: WideString = ''): Boolean;
var
  Index: Integer;
begin
  Result := False;

  if InternalFind(SectionName, Index) then
  begin
    if KeyName <> '' then
      Result := TXmlIniSection(FSortedList[Index]).Exists(KeyName)
    else
      Result := True;
  end;
end;

procedure TXmlIni.Clear;
var
  i: Integer;
begin
  FXmlEncoding := xetUTF_16;
  FEncryptMode := xiemAuto;

  for i := Count - 1 downto 0 do
    SectionByIndex[i].Free;
end;

procedure TXmlIni.Collation;
var
  i: Integer;
begin
  for i := Count - 1 downto 0 do
  begin
    SectionByIndex[i].Collation;
    if SectionByIndex[i].Count = 0 then
      SectionByIndex[i].Free;
  end;
end;

procedure TXmlIni.LoadFromStream(Stream: TStream);
var
  TestBuf: array[0..1] of Char;
  UnzipMS: TMemoryStream;
  {$IFDEF AES_ENCRYPT_ENABLED}
  Key128: TAESKey128;
  Key256: TAESKey256;
  FullPassword: string;
  DecryptMS: TMemoryStream;
  procedure LoadAESDecrypt;
  begin
    DecryptMS := TMemoryStream.Create;
    try
      DecryptAESStreamECB(Stream, Stream.Size-2, Key128, DecryptMS);

      UnzipMS := TMemoryStream.Create;
      try
        DecryptMS.Position := 0;
        DecompressStream(DecryptMS, UnzipMS);
        InternalXmlLoadFromStream(UnzipMS);
      finally
        UnzipMS.Free;
      end;
    finally
      DecryptMS.Free;
    end;
  end;
  procedure LoadAESDecrypt_LV1;
  var
    RSize: Byte;
    RStr: string;
    p: Integer;
  begin
    DecryptMS := TMemoryStream.Create;
    try
      DecryptAESStreamECB(Stream, Stream.Size-2, Key128, DecryptMS);

      UnzipMS := TMemoryStream.Create;
      try
        DecryptMS.Position := 0;
        DecryptMS.Read(RSize, SizeOf(RSize));
        RStr := StringOfChar(' ', RSize);
        DecryptMS.Read(PChar(RStr)^, RSize);
        p := SizeOf(RSize)+RSize;
        while p < DecryptMS.Size do
        begin
          (PChar(DecryptMS.Memory)+p)^ := Char(Byte((PChar(DecryptMS.Memory)+p)^) xor Byte(RStr[(p mod RSize) + 1]));
          Inc(p);
        end;
        DecompressStream(DecryptMS, UnzipMS);
        InternalXmlLoadFromStream(UnzipMS);
      finally
        UnzipMS.Free;
      end;
    finally
      DecryptMS.Free;
    end;
  end;
  procedure LoadAESDecrypt_LV2;
  var
    RSize: Byte;
    RStr: string;
    p: Integer;
  begin
    DecryptMS := TMemoryStream.Create;
    try
      DecryptAESStreamECB(Stream, Stream.Size-2, Key256, DecryptMS);

      UnzipMS := TMemoryStream.Create;
      try
        DecryptMS.Position := 0;
        DecryptMS.Read(RSize, SizeOf(RSize));
        RStr := StringOfChar(' ', RSize);
        DecryptMS.Read(PChar(RStr)^, RSize);
        p := SizeOf(RSize)+RSize;
        while p < DecryptMS.Size do
        begin
          (PChar(DecryptMS.Memory)+p)^ := Char(Byte((PChar(DecryptMS.Memory)+p)^) xor Byte(RStr[(p mod RSize) + 1]));
          Inc(p);
        end;
        DecompressStream(DecryptMS, UnzipMS);
        InternalXmlLoadFromStream(UnzipMS);
      finally
        UnzipMS.Free;
      end;
    finally
      DecryptMS.Free;
    end;
  end;
  {$ENDIF}
  {$IFDEF RSA_ENCRYPT_ENABLED}
  procedure LoadRSACrypto_LV3;
  var
    RSize: Byte;
    RStr: string;
    p: Integer;
    SS: TStringStream;
    Signature: string;
    SignatureSize: Word;
  begin
    DecryptMS := TMemoryStream.Create;
    try
      DecryptAESStreamECB(Stream, Stream.Size-2, Key256, DecryptMS);

      UnzipMS := TMemoryStream.Create;
      try
        DecryptMS.Position := 0;
        DecryptMS.Read(RSize, SizeOf(RSize));
        RStr := StringOfChar(' ', RSize);
        DecryptMS.Read(PChar(RStr)^, RSize);

        DecryptMS.Read(SignatureSize, SizeOf(SignatureSize));
        Signature := StringOfChar(' ', SignatureSize);
        DecryptMS.Read(PChar(Signature)^, SignatureSize);

        p := SizeOf(RSize)+RSize+SizeOf(SignatureSize)+SignatureSize;
        while p < DecryptMS.Size do
        begin
          (PChar(DecryptMS.Memory)+p)^ := Char(Byte((PChar(DecryptMS.Memory)+p)^) xor Byte(RStr[(p mod RSize) + 1]));
          Inc(p);
        end;

        DecompressStream(DecryptMS, UnzipMS);

        SS := TStringStream.Create('');
        try
          UnzipMS.Position := 0;
          XmlT_MimeEncodeStream(UnzipMS, SS);
          if not RSAVerifySignStr(SS.DataString, Signature, FVerifySignPublicKey.Text) then
            Exit;
        finally
          SS.Free;
        end;

        InternalXmlLoadFromStream(UnzipMS);
      finally
        UnzipMS.Free;
      end;
    finally
      DecryptMS.Free;
    end;
  end;
  {$ENDIF}
{$IFDEF RSA_ENCRYPT_ENABLED}
var
  KeySize: Integer;
  RSize: Byte;
  RStr: string;
{$ENDIF}
begin
  try
    Stream.Position := 0;
    Stream.Read(TestBuf, 2);

    FEncryptMode := xiemPlain;

    {$IFDEF RSA_ENCRYPT_ENABLED}
    if (TestBuf = HeaderRSACrypto) or (TestBuf = HeaderRSACrypto_LV1) or
        (TestBuf = HeaderRSACrypto_LV2) or (TestBuf = HeaderRSACrypto_LV3) then
    begin
      if FPrivateKey.Text = '' then
        raise Exception.Create('Data encrypted, PrivateKey required.');
      if (TestBuf = HeaderRSACrypto_LV3) and (FVerifySignPublicKey.Text = '') then
        raise Exception.Create('Data signed, VerifySignPublicKey required.');

      FCompressed := True;
      Stream.Position := 2;

      if (TestBuf = HeaderRSACrypto_LV1) or (TestBuf = HeaderRSACrypto_LV2) or (TestBuf = HeaderRSACrypto_LV3) then
      begin
        Stream.Read(RSize, SizeOf(RSize));
        RStr := StringOfChar(' ', RSize);
        Stream.Read(PChar(RStr)^, RSize);
      end;

      Stream.Read(KeySize, SizeOf(KeySize));
      FullPassword := StringOfChar(' ', KeySize);
      Stream.Read(PChar(FullPassword)^, KeySize);

      FullPassword := RSADecryptStr(FullPassword, FPrivateKey.Text);

      FullPassword := Copy(FullPassword, FakePasswordLength+1, MaxInt);

      if (TestBuf = HeaderRSACrypto_LV2) or (TestBuf = HeaderRSACrypto_LV3) then
      begin
        FillChar(Key256, SizeOf(Key256), 0);
        Move(PChar(FullPassword)^, Key256, SizeOf(Key256));
      end
      else
      begin
        FillChar(Key128, SizeOf(Key128), 0);
        Move(PChar(FullPassword)^, Key128, SizeOf(Key128));
      end;

      if TestBuf = HeaderRSACrypto_LV3 then
      begin
        LoadRSACrypto_LV3;
        FEncryptMode := xiemRSA_E3;
      end
      else if TestBuf = HeaderRSACrypto_LV2 then
      begin
        LoadAESDecrypt_LV2;
        FEncryptMode := xiemRSA_E2;
      end
      else if TestBuf = HeaderRSACrypto_LV1 then
      begin
        LoadAESDecrypt_LV1;
        FEncryptMode := xiemRSA_E1;
      end
      else
      begin
        LoadAESDecrypt;
        FEncryptMode := xiemRSA;
      end;

      Exit;
    end;
    {$ENDIF}

    {$IFDEF AES_ENCRYPT_ENABLED}
    if (TestBuf = HeaderAESCrypto) or (TestBuf = HeaderAESCrypto_LV1) or
        (TestBuf = HeaderAESCrypto_LV2) then
    begin
      if FPassword = '' then
        raise Exception.Create('Data encrypted, Password required.');

      FCompressed := True;
      Stream.Position := 2;

      if TestBuf = HeaderAESCrypto_LV2 then
      begin
        FullPassword := FPassword;
        while Length(FullPassword) < SizeOf(Key256) do
          FullPassword := FullPassword + FPassword;

        FillChar(Key256, SizeOf(Key256), 0);
        Move(PChar(FullPassword)^, Key256, SizeOf(Key256));
      end
      else
      begin
        FullPassword := FPassword;
        while Length(FullPassword) < SizeOf(Key128) do
          FullPassword := FullPassword + FPassword;

        FillChar(Key128, SizeOf(Key128), 0);
        Move(PChar(FullPassword)^, Key128, SizeOf(Key128));
      end;

      if TestBuf = HeaderAESCrypto_LV2 then
      begin
        LoadAESDecrypt_LV2;
        FEncryptMode := xiemAES_E2;
      end
      else if TestBuf = HeaderAESCrypto_LV1 then
      begin
        LoadAESDecrypt_LV1;
        FEncryptMode := xiemAES_E1;
      end
      else
      begin
        LoadAESDecrypt;
        FEncryptMode := xiemAES;
      end;

      Exit;
    end;
    {$ENDIF}

    if TestBuf = HeaderCompress then
    begin
      FCompressed := True;
      UnzipMS := TMemoryStream.Create;
      try
        Stream.Position := 0;
        DecompressStream(Stream, UnzipMS);
        InternalXmlLoadFromStream(UnzipMS);
      finally
        UnzipMS.Free;
      end;

      Exit;
    end;

    FCompressed := False;
    Stream.Position := 0;
    InternalXmlLoadFromStream(Stream);
  except
    if FRaiseException then
      raise;
  end;
end;

procedure TXmlIni.SaveToStream(Stream: TStream);
var
  ZipMS: TMemoryStream;
  CurEncMode: TXmlIniEncryptMode;
  {$IFDEF AES_ENCRYPT_ENABLED}
  Key128: TAESKey128;
  Key256: TAESKey256;
  FullPassword: string;
  EncryptMS: TMemoryStream;
  procedure SaveAESEncrypt;
  begin
    EncryptMS := TMemoryStream.Create;
    try
      ZipMS := TMemoryStream.Create;
      try
        InternalXmlSaveToStream(ZipMS);
        ZipMS.Position := 0;
        EncryptMS.Position := 0;
        CompressStream(ZipMS, EncryptMS);
      finally
        ZipMS.Free;
      end;

      EncryptMS.Position := 0;
      EncryptAESStreamECB(EncryptMS, 0, Key128, Stream);
    finally
      EncryptMS.Free;
    end;
  end;
  procedure SaveAESEncrypt_LV1;
  var
    RSize: Byte;
    RStr: string;
    p: Integer;
  begin
    EncryptMS := TMemoryStream.Create;
    try
      ZipMS := TMemoryStream.Create;
      try
        InternalXmlSaveToStream(ZipMS);

        ZipMS.Position := 0;
        EncryptMS.Position := 0;

        RSize := 5+Random(5);
        RStr := RandomStr256Base(RSize);
        EncryptMS.Write(RSize, SizeOf(RSize));
        EncryptMS.Write(PChar(RStr)^, RSize);

        CompressStream(ZipMS, EncryptMS);

        p := SizeOf(RSize)+RSize;
        while p < EncryptMS.Size do
        begin
          (PChar(EncryptMS.Memory)+p)^ := Char(Byte((PChar(EncryptMS.Memory)+p)^) xor Byte(RStr[(p mod RSize) + 1]));
          Inc(p);
        end;
      finally
        ZipMS.Free;
      end;

      EncryptMS.Position := 0;
      EncryptAESStreamECB(EncryptMS, 0, Key128, Stream);
    finally
      EncryptMS.Free;
    end;
  end;
  procedure SaveAESEncrypt_LV2;
  var
    RSize: Byte;
    RStr: string;
    p: Integer;
  begin
    EncryptMS := TMemoryStream.Create;
    try
      ZipMS := TMemoryStream.Create;
      try
        InternalXmlSaveToStream(ZipMS);

        ZipMS.Position := 0;
        EncryptMS.Position := 0;

        RSize := 5+Random(5);
        RStr := RandomStr256Base(RSize);
        EncryptMS.Write(RSize, SizeOf(RSize));
        EncryptMS.Write(PChar(RStr)^, RSize);

        CompressStream(ZipMS, EncryptMS);

        p := SizeOf(RSize)+RSize;
        while p < EncryptMS.Size do
        begin
          (PChar(EncryptMS.Memory)+p)^ := Char(Byte((PChar(EncryptMS.Memory)+p)^) xor Byte(RStr[(p mod RSize) + 1]));
          Inc(p);
        end;
      finally
        ZipMS.Free;
      end;

      EncryptMS.Position := 0;
      EncryptAESStreamECB(EncryptMS, 0, Key256, Stream);
    finally
      EncryptMS.Free;
    end;
  end;
  {$ENDIF}
  {$IFDEF RSA_ENCRYPT_ENABLED}
  procedure SaveRSAEncrypt_LV3;
  var
    RSize: Byte;
    RStr: string;
    p: Integer;
    SS: TStringStream;
    Signature: string;
    SignatureSize: Word;
  begin
    EncryptMS := TMemoryStream.Create;
    try
      ZipMS := TMemoryStream.Create;
      try
        InternalXmlSaveToStream(ZipMS);

        SS := TStringStream.Create('');
        try
          ZipMS.Position := 0;
          XmlT_MimeEncodeStream(ZipMS, SS);
          Signature := RSASignStr(SS.DataString, FSignPrivateKey.Text);
        finally
          SS.Free;
        end;

        ZipMS.Position := 0;
        EncryptMS.Position := 0;

        RSize := 5+Random(5);
        RStr := RandomStr256Base(RSize);
        EncryptMS.Write(RSize, SizeOf(RSize));
        EncryptMS.Write(PChar(RStr)^, RSize);

        SignatureSize := Length(Signature);
        EncryptMS.Write(SignatureSize, SizeOf(SignatureSize));
        EncryptMS.Write(PChar(Signature)^, SignatureSize);

        CompressStream(ZipMS, EncryptMS);

        p := SizeOf(RSize)+RSize+SizeOf(SignatureSize)+SignatureSize;
        while p < EncryptMS.Size do
        begin
          (PChar(EncryptMS.Memory)+p)^ := Char(Byte((PChar(EncryptMS.Memory)+p)^) xor Byte(RStr[(p mod RSize) + 1]));
          Inc(p);
        end;
      finally
        ZipMS.Free;
      end;

      EncryptMS.Position := 0;
      EncryptAESStreamECB(EncryptMS, 0, Key256, Stream);
    finally
      EncryptMS.Free;
    end;
  end;
  {$ENDIF}
{$IFDEF RSA_ENCRYPT_ENABLED}
var
  KeySize: Integer;
  RSize: Byte;
{$ENDIF}
begin
  try
    CurEncMode := FEncryptMode;

{$IFDEF RSA_ENCRYPT_ENABLED}
    if CurEncMode in [xiemRSA, xiemRSA_E1, xiemRSA_E2, xiemRSA_E3] then
    begin
      if FPublicKey.Text = '' then
        if RaiseException then
          raise Exception.Create('RSA encrypt, PublicKey required.')
        else
          CurEncMode := xiemAuto;
      if CurEncMode = xiemRSA_E3 then
        if FVerifySignPublicKey.Text = '' then
          if RaiseException then
            raise Exception.Create('RSA sign, VerifySignPublicKey required.')
          else
            CurEncMode := xiemAuto;
    end;
{$ENDIF}
{$IFDEF AES_ENCRYPT_ENABLED}
    if CurEncMode in [xiemAES, xiemAES_E1, xiemAES_E2] then
    begin
      if FPassword = '' then
        if RaiseException then
          raise Exception.Create('AES encrypt, Password required.')
        else
          CurEncMode := xiemAuto;
    end;
{$ENDIF}

    if CurEncMode = xiemAuto then
    begin
{$IFDEF RSA_ENCRYPT_ENABLED}
      if FPublicKey.Text <> '' then
      begin
        if FSignPrivateKey.Text <> '' then
          CurEncMode := xiemRSA_E3
        else
          CurEncMode := xiemRSA_E2;
      end
      else
{$ENDIF}
{$IFDEF AES_ENCRYPT_ENABLED}
      if FPassword <> '' then
        CurEncMode := xiemAES_E1
      else
{$ENDIF}
        CurEncMode := xiemPlain;
    end;

    case CurEncMode of
{$IFDEF RSA_ENCRYPT_ENABLED}
      xiemRSA, xiemRSA_E1, xiemRSA_E2, xiemRSA_E3:
        begin
          FCompressed := True;

          if (CurEncMode = xiemRSA_E2) or (CurEncMode = xiemRSA_E3) then
          begin
            FullPassword := RandomStr256Base(SizeOf(Key256));
            Move(PChar(FullPassword)^, Key256, SizeOf(Key256));
          end
          else
          begin
            FullPassword := RandomStr256Base(SizeOf(Key128));
            Move(PChar(FullPassword)^, Key128, SizeOf(Key128));
          end;

          FullPassword := RandomStr256Base(FakePasswordLength) + FullPassword;

          FullPassword := RSAEncryptStr(FullPassword, FPublicKey.Text);
          KeySize := Length(FullPassword);

          Stream.Position := 0;

          if CurEncMode = xiemRSA_E3 then
          begin
            Stream.Write(HeaderRSACrypto_LV3, 2);
            RSize := 5+Random(5);
            Stream.Write(RSize, SizeOf(RSize));
            Stream.Write(PChar(RandomStr256Base(RSize))^, RSize);
            Stream.Write(KeySize, SizeOf(KeySize));
            Stream.Write(PChar(FullPassword)^, KeySize);
            SaveRSAEncrypt_LV3;
          end
          else if CurEncMode = xiemRSA_E2 then
          begin
            Stream.Write(HeaderRSACrypto_LV2, 2);
            RSize := 5+Random(5);
            Stream.Write(RSize, SizeOf(RSize));
            Stream.Write(PChar(RandomStr256Base(RSize))^, RSize);
            Stream.Write(KeySize, SizeOf(KeySize));
            Stream.Write(PChar(FullPassword)^, KeySize);
            SaveAESEncrypt_LV2;
          end
          else if CurEncMode = xiemRSA_E1 then
          begin
            Stream.Write(HeaderRSACrypto_LV1, 2);
            RSize := 5+Random(5);
            Stream.Write(RSize, SizeOf(RSize));
            Stream.Write(PChar(RandomStr256Base(RSize))^, RSize);
            Stream.Write(KeySize, SizeOf(KeySize));
            Stream.Write(PChar(FullPassword)^, KeySize);
            SaveAESEncrypt_LV1;
          end
          else
          begin
            Stream.Write(HeaderRSACrypto, 2);
            Stream.Write(KeySize, SizeOf(KeySize));
            Stream.Write(PChar(FullPassword)^, KeySize);
            SaveAESEncrypt;
          end;
        end;
{$ENDIF}
{$IFDEF AES_ENCRYPT_ENABLED}
      xiemAES, xiemAES_E1, xiemAES_E2:
        begin
          FCompressed := True;

          if CurEncMode = xiemAES_E2 then
          begin
            FullPassword := FPassword;
            while Length(FullPassword) < SizeOf(Key256) do
              FullPassword := FullPassword + FPassword;

            FillChar(Key256, SizeOf(Key256), 0);
            Move(PChar(FullPassword)^, Key256, SizeOf(Key256));
          end
          else
          begin
            FullPassword := FPassword;
            while Length(FullPassword) < SizeOf(Key128) do
              FullPassword := FullPassword + FPassword;

            FillChar(Key128, SizeOf(Key128), 0);
            Move(PChar(FullPassword)^, Key128, SizeOf(Key128));
          end;

          Stream.Position := 0;
          if CurEncMode = xiemAES_E2 then
          begin
            Stream.Write(HeaderAESCrypto_LV2, 2);
            SaveAESEncrypt_LV2;
          end
          else if CurEncMode = xiemAES_E1 then
          begin
            Stream.Write(HeaderAESCrypto_LV1, 2);
            SaveAESEncrypt_LV1;
          end
          else
          begin
            Stream.Write(HeaderAESCrypto, 2);
            SaveAESEncrypt;
          end;
        end;
{$ENDIF}
      xiemPlain:
        begin
          if FCompressed then
          begin
            ZipMS := TMemoryStream.Create;
            try
              InternalXmlSaveToStream(ZipMS);
              ZipMS.Position := 0;
              Stream.Position := 0;
              CompressStream(ZipMS, Stream);
            finally
              ZipMS.Free;
            end;
          end
          else
            InternalXmlSaveToStream(Stream);
        end;
    end;
  except
    if FRaiseException then
      raise;
  end;
end;

procedure TXmlIni.LoadFromString(Value: WideString);
var
  SS: TStringStream;
begin
  if TrimLeft(Value) = '' then
    Clear
  else if TrimLeft(Value)[1] = '<' then
  begin
    SS := TStringStream.Create(Value);
    try
      LoadFromStream(SS);
    finally
      SS.Free;
    end;
  end
  else
    LoadFromBase64String(Value);
end;

function TXmlIni.SaveToString: WideString;
var
  OldXIEncoding: TXmlEncodingType;
  MS: TMemoryStream;
  NullChar: WideChar;
begin
  if not FCompressed then
  begin
    OldXIEncoding := XmlEncoding;
    try
      XmlEncoding := xetUTF_16;

      MS := TMemoryStream.Create;
      try
        SaveToStream(MS);
        NullChar := #0;
        MS.Write(NullChar, 2);
        Result := PWideChar(PChar(MS.Memory) + 2);
      finally
        MS.Free;
      end;
    finally
      XmlEncoding := OldXIEncoding;
    end;
  end
  else
    Result := SaveToBase64String;
end;

procedure TXmlIni.LoadFromBase64String(Value: string);
var
  SS: TStringStream;
  MS: TMemoryStream;
begin
  if (TrimLeft(Value) = '') or (TrimLeft(Value)[1] = '<') then
    LoadFromString(Value)
  else
  begin
    SS := TStringStream.Create(Trim(Value));
    try
      MS := TMemoryStream.Create;
      try
        SS.Position := 0;
        XmlT_MimeDecodeStream(SS, MS);
        LoadFromStream(MS);
      finally
        MS.Free;
      end;
    finally
      SS.Free;
    end;
  end;
end;

function TXmlIni.SaveToBase64String: string;
var
  SS: TStringStream;
  MS: TMemoryStream;
begin
  MS := TMemoryStream.Create;
  try
    SaveToStream(MS);

    SS := TStringStream.Create('');
    try
      MS.Position := 0;
      XmlT_MimeEncodeStream(MS, SS);
      Result := SS.DataString;
    finally
      SS.Free;
    end;
  finally
    MS.Free;
  end;
end;

procedure TXmlIni.LoadFromFile(AFileName: WideString = '');
begin
  AFileName := Trim(AFileName);

  if AFileName = '' then
  begin
    if FFileName <> '' then
      AFileName := FFileName
    else if FGlobalID <> '' then
      AFileName := FGlobalID
    else
      AFileName := ExtractFileNameWithoutExt(ParamStr(0)) + '.xi';
  end;

  if (AFileName <> '') and (ExtractFilePath(AFileName) = '') then
    AFileName := ExtractFilePath(ParamStr(0)) + AFileName;

  if not FileExists(AFileName) then
    if FRaiseException then
      raise Exception.Create('File "' + AFileName + '" not exists.')
    else
      Exit;

  FFileName := AFileName;

  InternalLoadFromFile(FFileName);
end;

procedure TXmlIni.SaveToFile(AFileName: WideString = '');
begin
  AFileName := Trim(AFileName);

  if AFileName = '' then
  begin
    if FFileName <> '' then
      AFileName := FFileName
    else if FGlobalID <> '' then
      AFileName := FGlobalID
    else
      AFileName := ExtractFileNameWithoutExt(ParamStr(0)) + '.xi';
  end;

  if (AFileName <> '') and (ExtractFilePath(AFileName) = '') then
    AFileName := ExtractFilePath(ParamStr(0)) + AFileName;

  if AFileName = '' then
    if FRaiseException then
      raise Exception.Create('File name is required.')
    else
      Exit;

  FFileName := AFileName;

  InternalSaveToFile(FFileName);
end;

procedure TXmlIni.UpdateToFile(AFileName: WideString = '');
var
  NowXI: TXmlIni;
begin
  AFileName := Trim(AFileName);

  if AFileName = '' then
  begin
    if FFileName <> '' then
      AFileName := FFileName
    else if FGlobalID <> '' then
      AFileName := FGlobalID
    else
      AFileName := ExtractFileNameWithoutExt(ParamStr(0)) + '.xi';
  end;

  if (AFileName <> '') and (ExtractFilePath(AFileName) = '') then
    AFileName := ExtractFilePath(ParamStr(0)) + AFileName;

  if AFileName = '' then
    if FRaiseException then
      raise Exception.Create('File name is required.')
    else
      Exit;

  NowXI := TXmlIni.Create;
  try
    NowXI.LoadFromFile(AFileName);
    NowXI.Append(Self);
    if not Self.Equals(NowXI) then
      NowXI.SaveToFile(AFileName);
  finally
    NowXI.Free;
  end;
end;

procedure TXmlIni.LoadFromRegistry(Path: WideString = ''; RootKey: HKEY = HKEY_LOCAL_MACHINE);
var
  Reg: TRegistry;
  Names: TStringList;
  i: Integer;
begin
  Clear;

  Path := Trim(Path);

  if Path = '' then
  begin
    if FGlobalID <> '' then
      Path := FGlobalID
    else
      Path := ExtractFileNameWithoutExt(ParamStr(0));
  end;

  ExtractRegPath(Path, RootKey);

  Reg := TRegistry.Create;
  try
    Reg.RootKey := RootKey;

    if Reg.OpenKey(Path, False) then
    begin
      Names := TStringList.Create;
      try
        Reg.GetKeyNames(Names);
        for i := 0 to Names.Count - 1 do
          Section[Names[i]].LoadFromRegistry(Path, RootKey, Names[i]);

        Section[''].LoadFromRegistry(Path, RootKey);

        Collation;
      finally
        Names.Free;
      end;
    end;
  finally
    Reg.Free;
  end;
end;

procedure TXmlIni.SaveToRegistry(Path: WideString = ''; RootKey: HKEY = HKEY_LOCAL_MACHINE);
var
  Reg: TRegistry;
  Names: TStringList;
  i: Integer;
begin
  Collation;

  Path := Trim(Path);

  if Path = '' then
  begin
    if FGlobalID <> '' then
      Path := FGlobalID
    else
      Path := ExtractFileNameWithoutExt(ParamStr(0));
  end;

  ExtractRegPath(Path, RootKey);

  Reg := TRegistry.Create;
  try
    Reg.RootKey := RootKey;

    if Reg.OpenKey(Path, False) then
    begin
      Names := TStringList.Create;
      try
        Reg.GetKeyNames(Names);
        for i := 0 to Names.Count - 1 do
          Reg.DeleteKey(Names[i])
      finally
        Names.Free;
      end;
    end;
  finally
    Reg.Free;
  end;

  for i := 0 to Count - 1 do
    SectionByIndex[i].SaveToRegistry(Path, RootKey, SectionByIndex[i].Name);
end;

procedure TXmlIni.UpdateToRegistry(Path: WideString = ''; RootKey: HKEY = HKEY_LOCAL_MACHINE);
var
  i: Integer;
begin
  Collation;

  Path := Trim(Path);

  if Path = '' then
  begin
    if FGlobalID <> '' then
      Path := FGlobalID
    else
      Path := ExtractFileNameWithoutExt(ParamStr(0));
  end;

  ExtractRegPath(Path, RootKey);

  for i := 0 to Count - 1 do
    SectionByIndex[i].UpdateToRegistry(Path, RootKey, SectionByIndex[i].Name);
end;

procedure TXmlIni.LoadFromIni(AIniFileName: WideString = '');
var
  Ini: TIniFile;
  SectionNames, KeyNameValues: TStringList;
  SecName, KeyName, KeyValue: WideString;
  XISection: TXmlIniSection;
  i, j: Integer;
begin
  Clear;

  AIniFileName := Trim(AIniFileName);

  if AIniFileName = '' then
  begin
    if FGlobalID <> '' then
      AIniFileName := FGlobalID
    else
      AIniFileName := ExtractFileNameWithoutExt(ParamStr(0)) + '.ini';
  end;

  if (AIniFileName <> '') and (ExtractFilePath(AIniFileName) = '') then
    AIniFileName := ExtractFilePath(ParamStr(0)) + AIniFileName;

  if (AIniFileName = '') or not FileExists(AIniFileName) then
    if FRaiseException then
      raise Exception.Create('File "' + AIniFileName + '" not exists.')
    else
      Exit;

  Ini := TIniFile.Create(AIniFileName);
  try
    SectionNames := TStringList.Create;
    KeyNameValues := TStringList.Create;
    try
      SectionNames.Duplicates := dupIgnore;
      KeyNameValues.Duplicates := dupIgnore;

      Ini.ReadSections(SectionNames);

      for i := 0 to SectionNames.Count - 1 do
      begin
        SecName := SingleLineToText(SectionNames[i]);
        XISection := InternalAdd(SecName);

        Ini.ReadSectionValues(SectionNames[i], KeyNameValues);

        for j := 0 to KeyNameValues.Count - 1 do
        begin
          KeyValue := KeyNameValues.ValueFromIndex[j];
          if KeyValue <> '' then
          begin
            KeyName := SingleLineToText(KeyNameValues.Names[j]);
            XISection.InternalAdd(KeyName).Text := SingleLineToText(KeyValue);
          end;
        end;
      end;

      Collation;
    finally
      KeyNameValues.Free;
      SectionNames.Free;
    end;
  finally
    Ini.Free;
  end;
end;

procedure TXmlIni.SaveToIni(AIniFileName: WideString = '');
var
  Ini: TIniFile;
  i, j: Integer;
begin
  Collation;

  AIniFileName := Trim(AIniFileName);

  if AIniFileName = '' then
  begin
    if FGlobalID <> '' then
      AIniFileName := FGlobalID
    else
      AIniFileName := ExtractFileNameWithoutExt(ParamStr(0)) + '.ini';
  end;

  if (AIniFileName <> '') and (ExtractFilePath(AIniFileName) = '') then
    AIniFileName := ExtractFilePath(ParamStr(0)) + AIniFileName;

  if AIniFileName = '' then
    if FRaiseException then
      raise Exception.Create('File name is required.')
    else
      Exit;

  DeleteFile(AIniFileName);

  Ini := TIniFile.Create(AIniFileName);
  try
    for i := 0 to Count - 1 do
      for j := 0 to SectionByIndex[i].Count - 1 do
        Ini.WriteString(
          TextToSingleLine(SectionByIndex[i].Name),
          TextToSingleLine(SectionByIndex[i].KeyByIndex[j].Name),
          TextToSingleLine(SectionByIndex[i].KeyByIndex[j].Text));
  finally
    Ini.Free;
  end;
end;

procedure TXmlIni.UpdateToIni(AIniFileName: WideString = '');
var
  Ini: TIniFile;
  i, j: Integer;
begin
  Collation;

  AIniFileName := Trim(AIniFileName);

  if AIniFileName = '' then
  begin
    if FGlobalID <> '' then
      AIniFileName := FGlobalID
    else
      AIniFileName := ExtractFileNameWithoutExt(ParamStr(0)) + '.ini';
  end;

  if (AIniFileName <> '') and (ExtractFilePath(AIniFileName) = '') then
    AIniFileName := ExtractFilePath(ParamStr(0)) + AIniFileName;

  if AIniFileName = '' then
    if FRaiseException then
      raise Exception.Create('File name is required.')
    else
      Exit;

  Ini := TIniFile.Create(AIniFileName);
  try
    for i := 0 to Count - 1 do
      for j := 0 to SectionByIndex[i].Count - 1 do
        Ini.WriteString(
          TextToSingleLine(SectionByIndex[i].Name),
          TextToSingleLine(SectionByIndex[i].KeyByIndex[j].Name),
          TextToSingleLine(SectionByIndex[i].KeyByIndex[j].Text));
  finally
    Ini.Free;
  end;
end;

type
  TSetMemMemoryStream = class(TMemoryStream)
  public
    procedure SetMem(Ptr: Pointer; Size: Longint);
  end;

procedure TSetMemMemoryStream.SetMem(Ptr: Pointer; Size: Integer);
begin
  SetPointer(Ptr, Size);
end;

procedure TXmlIni.LoadFromShareMem(ShareMemID: string = '');
var
  MemLockMutex: TMutex;
  MS: TSetMemMemoryStream;
  ShareMemHandle: THandle;
  ShareMemMap: Pointer;
  ShareMemSize: Integer;
begin
  ShareMemID := Trim(ShareMemID);

  if ShareMemID = '' then
  begin
    if FGlobalID <> '' then
      ShareMemID := FGlobalID
    else
      ShareMemID := ExtractFileNameWithoutExt(ParamStr(0));
  end;

  ShareMemID := UpperCase(Trim(ShareMemID));

  if ShareMemID = '' then
    if FRaiseException then
      raise Exception.Create('ShareMemID is required.')
    else
      Exit;

  ShareMemID := XI_ShareName + ShareMemID;

  MemLockMutex := TMutex.Create(XI_LockMutexName + ShareMemID);
  try
    if MemLockMutex.Get then
    begin
      try
        try
          ShareMemHandle := CreateFileMapping($FFFFFFFF, nil, PAGE_READONLY, 0,
            SizeOf(Integer), PChar(ShareMemID));

          ShareMemMap := MapViewOfFile(ShareMemHandle, FILE_MAP_READ, 0, 0, SizeOf(Integer));
          try
            Move(ShareMemMap^, ShareMemSize, SizeOf(Integer));
          finally
            UnmapViewOfFile(ShareMemMap);
          end;

          if ShareMemSize = 0 then
            Clear
          else
          begin
            ShareMemHandle := CreateFileMapping($FFFFFFFF, nil, PAGE_READONLY, 0,
              SizeOf(Integer) + ShareMemSize, PChar(ShareMemID));

            ShareMemMap := MapViewOfFile(ShareMemHandle, FILE_MAP_READ, 0, 0, SizeOf(Integer) + ShareMemSize);
            try
              MS := TSetMemMemoryStream.Create;
              try
                MS.SetMem(Pointer(PChar(ShareMemMap) + SizeOf(Integer)), ShareMemSize);
                LoadFromStream(MS);
                MS.SetMem(nil, 0);
              finally
                MS.Free;
              end;
            finally
              UnmapViewOfFile(ShareMemMap);
            end;
          end;
        except
          on Exception do
            if FRaiseException then
              raise;
        end;
      finally
        MemLockMutex.Release;
      end;
    end;
  finally
    MemLockMutex.Free;
  end;
end;

procedure TXmlIni.SaveToShareMem(ShareMemID: string = '');
var
  MemLockMutex: TMutex;
  MS: TMemoryStream;
  ShareMemHandle: THandle;
  ShareMemMap: Pointer;
  ShareMemSize: Integer;
begin
  ShareMemID := Trim(ShareMemID);

  if ShareMemID = '' then
  begin
    if FGlobalID <> '' then
      ShareMemID := FGlobalID
    else
      ShareMemID := ExtractFileNameWithoutExt(ParamStr(0));
  end;

  ShareMemID := UpperCase(Trim(ShareMemID));

  if ShareMemID = '' then
    if FRaiseException then
      raise Exception.Create('ShareMemID is required.')
    else
      Exit;

  ShareMemID := XI_ShareName + ShareMemID;

  MemLockMutex := TMutex.Create(XI_LockMutexName + ShareMemID);
  try
    if MemLockMutex.Get then
    begin
      try
        try
          MS := TMemoryStream.Create;
          try
            SaveToStream(MS);

            ShareMemHandle := CreateFileMapping($FFFFFFFF, nil, PAGE_READWRITE, 0,
              SizeOf(Integer) + MS.Size, PChar(ShareMemID));

            ShareMemMap := MapViewOfFile(ShareMemHandle, FILE_MAP_WRITE, 0, 0, SizeOf(Integer) + MS.Size);
            try
              ShareMemSize := MS.Size;
              Move(ShareMemSize, ShareMemMap^, SizeOf(Integer));
              Move(MS.Memory^, Pointer(PChar(ShareMemMap) + SizeOf(Integer))^, MS.Size);
            finally
              UnmapViewOfFile(ShareMemMap);
            end;
          finally
            MS.Free;
          end;
        except
          on Exception do
            if FRaiseException then
              raise;
        end;
      finally
        MemLockMutex.Release;
      end;
    end;
  finally
    MemLockMutex.Free;
  end;
end;

procedure TXmlIni.UpdateToShareMem(ShareMemID: string = '');
var
  MemLockMutex: TMutex;
  SMS: TSetMemMemoryStream;
  MS: TMemoryStream;
  ShareMemHandle: THandle;
  ShareMemMap: Pointer;
  ShareMemSize: Integer;
  NowXI: TXmlIni;
begin
  ShareMemID := Trim(ShareMemID);

  if ShareMemID = '' then
  begin
    if FGlobalID <> '' then
      ShareMemID := FGlobalID
    else
      ShareMemID := ExtractFileNameWithoutExt(ParamStr(0));
  end;

  ShareMemID := UpperCase(Trim(ShareMemID));

  if ShareMemID = '' then
    if FRaiseException then
      raise Exception.Create('ShareMemID is required.')
    else
      Exit;

  ShareMemID := XI_ShareName + ShareMemID;

  NowXI := TXmlIni.Create;
  MemLockMutex := TMutex.Create(XI_LockMutexName + ShareMemID);
  try
    if MemLockMutex.Get then
    begin
      try
        try
          ShareMemHandle := CreateFileMapping($FFFFFFFF, nil, PAGE_READONLY, 0,
            SizeOf(Integer), PChar(ShareMemID));

          ShareMemMap := MapViewOfFile(ShareMemHandle, FILE_MAP_READ, 0, 0, SizeOf(Integer));
          try
            Move(ShareMemMap^, ShareMemSize, SizeOf(Integer));
          finally
            UnmapViewOfFile(ShareMemMap);
          end;

          if ShareMemSize = 0 then
            NowXI.Clear
          else
          begin
            ShareMemHandle := CreateFileMapping($FFFFFFFF, nil, PAGE_READONLY, 0,
              SizeOf(Integer) + ShareMemSize, PChar(ShareMemID));

            ShareMemMap := MapViewOfFile(ShareMemHandle, FILE_MAP_READ, 0, 0, SizeOf(Integer) + ShareMemSize);
            try
              SMS := TSetMemMemoryStream.Create;
              try
                SMS.SetMem(Pointer(PChar(ShareMemMap) + SizeOf(Integer)), ShareMemSize);
                NowXI.LoadFromStream(SMS);
                SMS.SetMem(nil, 0);
              finally
                SMS.Free;
              end;
            finally
              UnmapViewOfFile(ShareMemMap);
            end;
          end;
        except
          on Exception do
            if FRaiseException then
              raise;
        end;

        NowXI.Append(Self);

        try
          MS := TMemoryStream.Create;
          try
            NowXI.SaveToStream(MS);

            ShareMemHandle := CreateFileMapping($FFFFFFFF, nil, PAGE_READWRITE, 0,
              SizeOf(Integer) + MS.Size, PChar(ShareMemID));

            ShareMemMap := MapViewOfFile(ShareMemHandle, FILE_MAP_WRITE, 0, 0, SizeOf(Integer) + MS.Size);
            try
              ShareMemSize := MS.Size;
              Move(ShareMemSize, ShareMemMap^, SizeOf(Integer));
              Move(MS.Memory^, Pointer(PChar(ShareMemMap) + SizeOf(Integer))^, MS.Size);
            finally
              UnmapViewOfFile(ShareMemMap);
            end;
          finally
            MS.Free;
          end;
        except
          on Exception do
            if FRaiseException then
              raise;
        end;
      finally
        MemLockMutex.Release;
      end;
    end;
  finally
    MemLockMutex.Free;
    NowXI.Free;
  end;
end;

procedure TXmlIni.LoadFromMedia(MediaType: TXmlIniMediaType; MediaID: string = '');
begin
  case MediaType of
    ximtFile:
      LoadFromFile(MediaID);
    ximtIni:
      LoadFromIni(MediaID);
    ximtBase64String:
      LoadFromBase64String(MediaID);
    ximtString:
      LoadFromString(MediaID);
    ximtRegistry:
      LoadFromRegistry(MediaID);
    ximtShareMem:
      LoadFromShareMem(MediaID);
  end;
end;

function TXmlIni.SaveToMedia(MediaType: TXmlIniMediaType; MediaID: string = ''): WideString;
begin
  Result := '';

  case MediaType of
    ximtFile:
      SaveToFile(MediaID);
    ximtIni:
      SaveToIni(MediaID);
    ximtBase64String:
      Result := SaveToBase64String;
    ximtString:
      Result := SaveToString;
    ximtRegistry:
      SaveToRegistry(MediaID);
    ximtShareMem:
      SaveToShareMem(MediaID);
  end;
end;

function TXmlIni.UpdateToMedia(MediaType: TXmlIniMediaType; MediaID: string = ''): WideString;
begin
  Result := '';

  case MediaType of
    ximtFile:
      UpdateToFile(MediaID);
    ximtIni:
      UpdateToIni(MediaID);
    ximtBase64String:
      Result := SaveToBase64String;
    ximtString:
      Result := SaveToString;
    ximtRegistry:
      UpdateToRegistry(MediaID);
    ximtShareMem:
      UpdateToShareMem(MediaID);
  end;
end;

{$IFDEF RSA_ENCRYPT_ENABLED}

procedure TXmlIni.GenerateNewKey;
begin
  GenerateNewEncryptKey;
  GenerateNewSignKey;
end;

procedure TXmlIni.GenerateNewEncryptKey;
var
  PubKey, PrvKey: string;
begin
  RSAGenKey(PubKey, PrvKey);
  FPublicKey.Text := PubKey;
  FPrivateKey.Text := PrvKey;
end;

procedure TXmlIni.GenerateNewSignKey;
var
  PubKey, PrvKey: string;
begin
  RSAGenKey(PubKey, PrvKey);
  FSignPrivateKey.Text := PrvKey;
  FVerifySignPublicKey.Text := PubKey;
end;

{$ENDIF}

procedure TXmlIni.Lock;
begin
  EnterCriticalSection(FLockCriSec);
end;

procedure TXmlIni.Unlock;
begin
  LeaveCriticalSection(FLockCriSec);
end;

function TXmlIni.SectionExists(const Section: WideString): Boolean;
begin
  Result := Exists(Section);
end;

function TXmlIni.ReadString(const Section, Ident: WideString; Default: WideString = ''): WideString;
begin
  if Self.Section[Section].Exists(Ident) then
    Result := Self.Section[Section][Ident]
  else
    Result := Default;
end;

procedure TXmlIni.WriteString(const Section, Ident: WideString; Value: WideString);
begin
  Self.Section[Section][Ident] := Value;
end;

function TXmlIni.ReadInteger(const Section, Ident: WideString; Default: Longint = 0): Longint;
begin
  if Self.Section[Section].Exists(Ident) then
    Result := Self.Section[Section].Key[Ident].AsInteger
  else
    Result := Default;
end;

procedure TXmlIni.WriteInteger(const Section, Ident: WideString; Value: Longint);
begin
  Self.Section[Section].Key[Ident].AsInteger := Value;
end;

function TXmlIni.ReadFloat(const Section, Ident: WideString; Default: Double = 0): Double;
begin
  if Self.Section[Section].Exists(Ident) then
    Result := Self.Section[Section].Key[Ident].AsFloat
  else
    Result := Default;
end;

procedure TXmlIni.WriteFloat(const Section, Ident: WideString; Value: Double);
begin
  Self.Section[Section].Key[Ident].AsFloat := Value;
end;

function TXmlIni.ReadBool(const Section, Ident: WideString; Default: Boolean = False): Boolean;
begin
  if Self.Section[Section].Exists(Ident) then
    Result := Self.Section[Section].Key[Ident].AsBool
  else
    Result := Default;
end;

procedure TXmlIni.WriteBool(const Section, Ident: WideString; Value: Boolean);
begin
  Self.Section[Section].Key[Ident].AsBool := Value;
end;

function TXmlIni.ReadDateTime(const Section, Ident: WideString; Default: TDateTime = 0): TDateTime;
begin
  if Self.Section[Section].Exists(Ident) then
    Result := Self.Section[Section].Key[Ident].AsDateTime
  else
    Result := Default;
end;

procedure TXmlIni.WriteDateTime(const Section, Ident: WideString; Value: TDateTime);
begin
  Self.Section[Section].Key[Ident].AsDateTime := Value;
end;

function TXmlIni.ReadDate(const Section, Ident: WideString; Default: TDateTime = 0): TDateTime;
begin
  if Self.Section[Section].Exists(Ident) then
    Result := Self.Section[Section].Key[Ident].AsDate
  else
    Result := Default;
end;

procedure TXmlIni.WriteDate(const Section, Ident: WideString; Value: TDateTime);
begin
  Self.Section[Section].Key[Ident].AsDate := Value;
end;

function TXmlIni.ReadTime(const Section, Ident: WideString; Default: TDateTime = 0): TDateTime;
begin
  if Self.Section[Section].Exists(Ident) then
    Result := Self.Section[Section].Key[Ident].AsTime
  else
    Result := Default;
end;

procedure TXmlIni.WriteTime(const Section, Ident: WideString; Value: TDateTime);
begin
  Self.Section[Section].Key[Ident].AsTime := Value;
end;

function TXmlIni.ReadBinaryStream(const Section, Ident: WideString; Value: TStream): Integer;
begin
  Result := Self.Section[Section].Key[Ident].AsStream.Size;
  if Result > 0 then
    Value.CopyFrom(Self.Section[Section].Key[Ident].AsStream, 0);
end;

procedure TXmlIni.WriteBinaryStream(const Section, Ident: WideString; Value: TStream);
begin
  Self.Section[Section].Key[Ident].AsStream := Value;
end;

procedure TXmlIni.ReadSection(const Section: WideString; Strings: TStrings);
begin
  ReadSectionValues(Section, Strings);
end;

procedure TXmlIni.ReadSections(Strings: TStrings);
var
  i: Integer;
begin
  Strings.BeginUpdate;
  try
    Strings.Clear;
    for i := 0 to Count - 1 do
      Strings.Add(SectionByIndex[i].Name);
  finally
    Strings.EndUpdate;
  end;
end;

procedure TXmlIni.ReadSectionValues(const Section: WideString; Strings: TStrings);
var
  i: Integer;
begin
  Strings.BeginUpdate;
  try
    Strings.Clear;
    if Exists(Section) then
      for i := 0 to Self.Section[Section].Count - 1 do
        Strings.Add(Self.Section[Section].KeyByIndex[i].Name + '=' + Self.Section[Section].KeyByIndex[i].Text);
  finally
    Strings.EndUpdate;
  end;
end;

procedure TXmlIni.EraseSection(const Section: WideString);
begin
  Self.Delete(Section);
end;

procedure TXmlIni.DeleteKey(const Section, Ident: WideString);
begin
  if SectionExists(Section) then
    Self.Section[Section].Key[Ident].Clear;
end;

function TXmlIni.ValueExists(const Section, Ident: WideString): Boolean;
begin
  if SectionExists(Section) then
    Result := Self.Section[Section].Exists(Ident)
  else
    Result := False;
end;

procedure TXmlIni.RegisterSection(Section: TXmlIniSection);
var
  Index: Integer;
begin
  Index := FList.IndexOf(Section);
  if Index < 0 then
    FList.Add(Section);

  if InternalFind(Section.FName, Index) then
    FSortedList.Delete(Index);

  InternalFind(Section.FName, Index);
  FSortedList.Insert(Index, Section);
end;

procedure TXmlIni.UnregisterSection(Section: TXmlIniSection);
var
  Index: Integer;
begin
  Index := FList.IndexOf(Section);
  if Index >= 0 then
    FList.Delete(Index);

  if InternalFind(Section.FName, Index) then
    FSortedList.Delete(Index);
end;

procedure TXmlIni.InternalXmlLoadFromStream(Stream: TStream);
var
  Root: IXmlNode;
  Enc: TXmlEncodingType;
  XMLDom: IXMLDocument;
  i: Integer;
  j: Integer;
begin
  CoInitialize(nil);

  CoInitialize(nil);

  XMLDom := NewXMLDocument;
  try
    Stream.Position := 0;
    XMLDom.LoadFromStream(Stream, xetUnknown);

    Clear;

    FXmlEncoding := xetUnknow;

    for Enc := Low(XT_XML_EncodingStr) to High(XT_XML_EncodingStr) do
      if XMLDom.Encoding = XT_XML_EncodingStr[Enc] then
      begin
        FXmlEncoding := Enc;
        Break;
      end;

    Root := XMLDom.DocumentElement;

    if Root.NodeName <> XI_Root_NodeName then
      Exit;

    for i := 0 to Root.ChildNodes.Count - 1 do
      if Root.ChildNodes[i].NodeName = XI_Section_NodeName then
        with Add(Root.ChildNodes[i].Attributes[XI_Name_Attribute]) do
        begin
          for j := 0 to Root.ChildNodes[i].ChildNodes.Count - 1 do
            if Root.ChildNodes[i].ChildNodes[j].NodeName = XI_Key_NodeName then
              Item[Root.ChildNodes[i].ChildNodes[j].Attributes[XI_Name_Attribute]] :=
                Root.ChildNodes[i].ChildNodes[j].Text;
        end;
  finally
    XMLDom := nil;
  end;

  Collation;
end;

procedure TXmlIni.InternalXmlSaveToStream(Stream: TStream);
  function GetLocalCodePage: Integer;
  const
    CP_ACP = 0;
    LOCALE_IDEFAULTANSICODEPAGE = $00001004;
  var
    ResultCode: Integer;
    Buffer: array[0..6] of Char;
  begin
    GetLocaleInfo(GetThreadLocale, LOCALE_IDEFAULTANSICODEPAGE, Buffer, SizeOf(Buffer));
    Val(Buffer, Result, ResultCode);
    if ResultCode <> 0 then
      Result := CP_ACP;
  end;
  function GetLocalEncodingType: TXmlEncodingType;
  begin
    case GetLocalCodePage of
      950: Result := xetBIG5;
      936: Result := xetGB2312;
      932: Result := xetS_JIS;
      949: Result := xetEUC_KR;
    else
      Result := xetUTF_8;
    end;
  end;
var
  Root, SectionNode, KeyNode: IXmlNode;
  XMLDom: IXMLDocument;
  i: Integer;
  j: Integer;
begin
  Collation;

  if FXmlEncoding = xetUnknow then
    FXmlEncoding := GetLocalEncodingType;

  CoInitialize(nil);

  XMLDom := NewXMLDocument;
  try
    XMLDom.Options := XMLDom.Options + [doNodeAutoIndent];
    XMLDom.NodeIndentStr := '  ';

    Root := XMLDom.AddChild(XI_ROOT_NodeName);

    for i := 0 to Count - 1 do
    begin
      SectionNode := Root.AddChild(XI_Section_NodeName);
      SectionNode.Attributes[XI_Name_Attribute] := SectionByIndex[i].Name;
      for j := 0 to SectionByIndex[i].Count - 1 do
      begin
        KeyNode := SectionNode.AddChild(XI_Key_NodeName);
        KeyNode.Attributes[XI_Name_Attribute] := SectionByIndex[i].KeyByIndex[j].Name;
        KeyNode.Text := SectionByIndex[i].KeyByIndex[j].Text;
      end;
    end;

    XMLDom.StandAlone := XI_XML_StandAlone;
    XMLDom.Version := XI_XML_Version;
    XMLDom.Encoding := XT_XML_EncodingStr[FXmlEncoding];

    Stream.Position := 0;
    if Stream is TStringStream then
      TStringStream(Stream).WriteString(XMLDom.XML.Text)
    else
      XMLDom.SaveToStream(Stream);
  finally
    XMLDom := nil
  end;
end;

procedure TXmlIni.InternalLoadFromFile(AFileName: WideString);
var
  MS: TMemoryStream;
  FileLockMutex: TMutex;
begin
  Clear;

  MS := TMemoryStream.Create;
  try
    FileLockMutex := TMutex.Create(
      StringReplace(XI_LockMutexName + AFileName, '\', '_', [rfReplaceAll, rfIgnoreCase]));
    try
      if FileLockMutex.Get then
      begin
        try
          MS.LoadFromFile(AFileName);
        finally
          FileLockMutex.Release;
        end;
      end;
    finally
      FileLockMutex.Free;
    end;

    LoadFromStream(MS);
  finally
    MS.Free;
  end;
end;

procedure TXmlIni.InternalSaveToFile(AFileName: WideString);
var
  MS: TMemoryStream;
  FileLockMutex: TMutex;
begin
  MS := TMemoryStream.Create;
  try
    SaveToStream(MS);

    FileLockMutex := TMutex.Create(
      StringReplace(XI_LockMutexName + AFileName, '\', '_', [rfReplaceAll, rfIgnoreCase]));
    try
      if FileLockMutex.Get then
      begin
        try
          MS.SaveToFile(AFileName);
        finally
          FileLockMutex.Release;
        end;
      end;
    finally
      FileLockMutex.Free;
    end;
  finally
    MS.Free;
  end;
end;

function TXmlIni.InternalFindCaseSensitive(const S: WideString; var Index: Integer): Boolean;
var
  L, H, I, C: Integer;
begin
  Result := False;
  L := 0;
  H := FSortedList.Count - 1;
  while L <= H do
  begin
    I := (L + H) shr 1;
    C := WideCompareStr(TXmlIniSection(FSortedList[I]).FName, S);
    if C < 0 then L := I + 1 else
    begin
      H := I - 1;
      if C = 0 then
      begin
        Result := True;
        L := I;
      end;
    end;
  end;
  Index := L;
end;

function TXmlIni.InternalFindCaseInsensitive(const S: WideString; var Index: Integer): Boolean;
var
  L, H, I, C: Integer;
begin
  Result := False;
  L := 0;
  H := FSortedList.Count - 1;
  while L <= H do
  begin
    I := (L + H) shr 1;
    C := WideCompareText(TXmlIniSection(FSortedList[I]).FName, S);
    if C < 0 then L := I + 1 else
    begin
      H := I - 1;
      if C = 0 then
      begin
        Result := True;
        L := I;
      end;
    end;
  end;
  Index := L;
end;

function TXmlIni.InternalFind(const SectionName: WideString; var Index: Integer): Boolean;
begin
  if FCaseSensitive then
    Result := InternalFindCaseSensitive(Trim(SectionName), Index)
  else
    Result := InternalFindCaseInsensitive(Trim(SectionName), Index);
end;

function TXmlIni.InternalAdd(SectionName: WideString): TXmlIniSection;
begin
  Result := TXmlIniSection.Create(SectionName, Self);
end;

function TXmlIni.GetCount: Integer;
begin
  Result := FList.Count;
end;

function TXmlIni.GetSectionByIndex(Index: Integer): TXmlIniSection;
begin
  if (Index >= 0) and (Index < FList.Count) then
    Result := TXmlIniSection(FList[Index])
  else
    Result := Add('');
end;

function TXmlIni.GetSection(Index: Variant): TXmlIniSection;
var
  IntIndex: Integer;
begin
  if VarIsNumeric(Index) then
  begin
    IntIndex := StrToIntDef(VarToWideStr(Index), 0);
    if (IntIndex >= 0) and (IntIndex < Count) then
      Result := TXmlIniSection(FList[IntIndex])
    else
      Result := Add('');
  end
  else
    Result := Add(VarToWideStr(Index));
end;

procedure TXmlIni.SetGlobalID(const Value: WideString);
begin
  if FGlobalID <> Value then
    FGlobalID := UpperCase(Trim(Value));
end;

procedure TXmlIni.SetFileName(const Value: WideString);
begin
  if FFileName <> Value then
  begin
    if FAutoSave then
      SaveToFile;
    if (Value <> '') and ((FGlobalID = '') or (FGlobalID = ExtractFileName(FFileName))) then
      FGlobalID := UpperCase(Trim(ExtractFileName(Value)));
    FFileName := Value;
    if FileExists(FFileName) then
      LoadFromFile;
  end;
end;

procedure TXmlIni.SetAutoSave(const Value: Boolean);
begin
  FAutoSave := Value;
  if FAutoSave then
    SaveToFile;
end;

procedure TXmlIni.SetCaseSensitive(const Value: Boolean);
var
  i: Integer;
begin
  if FCaseSensitive <> Value then
  begin
    FCaseSensitive := Value;

    FSortedList.Clear;
    for i := 0 to Count - 1 do
    begin
      RegisterSection(SectionByIndex[i]);
      SectionByIndex[i].CaseSensitive := FCaseSensitive;
    end;
  end;
end;

procedure TXmlIni.SetCompressed(const Value: Boolean);
var
  i: Integer;
begin
  if FCompressed <> Value then
  begin
    FCompressed := Value;

    if FAutoSave then
      SaveToFile;

    for i := 0 to Count - 1 do
      SectionByIndex[i].Compressed := FCompressed;
  end;
end;

procedure TXmlIni.SetEncryptMode(const Value: TXmlIniEncryptMode);
begin
  if FEncryptMode <> Value then
    FEncryptMode := Value;
end;

{$IFDEF AES_ENCRYPT_ENABLED}
procedure TXmlIni.SetPassword(const Value: WideString);
begin
  if FPassword <> Trim(Value) then
    FPassword := Trim(Value);
end;
{$ENDIF}

{$IFDEF RSA_ENCRYPT_ENABLED}
function TXmlIni.GetPrivateKey: TStrings;
begin
  Result := FPrivateKey;
end;

procedure TXmlIni.SetPrivateKey(const Value: TStrings);
begin
  FPrivateKey.Assign(Value);
end;

function TXmlIni.GetPublicKey: TStrings;
begin
  Result := FPublicKey;
end;

procedure TXmlIni.SetPublicKey(const Value: TStrings);
begin
  FSignPrivateKey.Assign(Value);
end;

function TXmlIni.GetSignPrivateKey: TStrings;
begin
  Result := FSignPrivateKey;
end;

procedure TXmlIni.SetSignPrivateKey(const Value: TStrings);
begin
  FPublicKey.Assign(Value);
end;

function TXmlIni.GetVerifySignPublicKey: TStrings;
begin
  Result := FVerifySignPublicKey;
end;

procedure TXmlIni.SetVerifySignPublicKey(const Value: TStrings);
begin
  FVerifySignPublicKey.Assign(Value);
end;
{$ENDIF}

procedure TXmlIni.SetRaiseException(const Value: Boolean);
var
  i: Integer;
begin
  if FRaiseException <> Value then
  begin
    FRaiseException := Value;
    for i := 0 to Count - 1 do
      SectionByIndex[i].RaiseException := FRaiseException;
  end;
end;

procedure TXmlIni.ExtractRegPath(var Path: WideString; var RootKey: HKEY);
const
  HKeyValues: array [0..4] of HKEY =
      (HKEY_CLASSES_ROOT, HKEY_CURRENT_USER, HKEY_LOCAL_MACHINE, HKEY_USERS, HKEY_CURRENT_CONFIG);
  HKeyStrs: array [0..4] of string =
      ('HKEY_CLASSES_ROOT\', 'HKEY_CURRENT_USER\', 'HKEY_LOCAL_MACHINE\', 'HKEY_USERS\', 'HKEY_CURRENT_CONFIG\');
var
  i: Integer;
  DetectPath: string;
begin
  Path := Trim(Path);

  while Pos('\\', Path) > 0 do
    Path := StringReplace(Path, '\\', '\', [rfReplaceAll]);

  if (Path = '') or (Path[Length(Path)] <> '\') then
    Path := Path + '\';

  if UpperCase(LeftStr(Path, 5)) = 'HKEY_' then
  begin
    DetectPath := UpperCase(Path);

    for i := Low(HKeyValues) to High(HKeyValues) do
      if LeftStr(DetectPath, Length(HKeyStrs[i])) = HKeyStrs[i] then
      begin
        Path := Copy(Path, Length(HKeyStrs[i])+1, MaxInt);
        RootKey := HKeyValues[i];
        Exit;
      end;
  end;

  if (Path = '') or (Path[1] <> '\') then
    Path := '\' + Path;
end;

{ TXmlIniSection }

constructor TXmlIniSection.Create(AName: WideString = ''; AOwner: TXmlIni = nil);
begin
  FOwner := AOwner;

  if FOwner <> nil then
    FRaiseException := FOwner.RaiseException
  else
    FRaiseException := False;

  FName := Trim(AName);

  FList := TList.Create;
  FSortedList := TList.Create;

  FXmlEncoding := xetUTF_16;
  FCaseSensitive := False;
  FCompressed := False;

  if FOwner <> nil then
  begin
    FOwner.RegisterSection(Self);
    FCaseSensitive := FOwner.FCaseSensitive;
    FCompressed := FOwner.FCompressed;
  end;
end;

destructor TXmlIniSection.Destroy;
begin
  Clear;

  if FOwner <> nil then
    FOwner.UnregisterSection(Self);

  FSortedList.Free;
  FList.Free;

  inherited;
end;

function TXmlIniSection.Equals(Dest: TXmlIniSection): Boolean;
var
  i: Integer;
begin
  Collation;
  Dest.Collation;
  Result := False;
  for i := 0 to Count - 1 do
  begin
    if not Dest.Exists(KeyByIndex[i].Name) then
      Exit;
    if Dest[KeyByIndex[i].Name] <> KeyByIndex[i].Text then
      Exit;
  end;
  Result := True;
end;

procedure TXmlIniSection.CopyFrom(Source: TXmlIniSection);
begin
  Clear;
  Append(Source);
end;

procedure TXmlIniSection.Append(Source: TXmlIniSection);
var
  i: Integer;
begin
  for i := 0 to Source.Count - 1 do
    SetItem(Source.KeyByIndex[i].Name, Source.KeyByIndex[i].Text);
end;

function TXmlIniSection.Equals(Dest: TStrings): Boolean;
var
  i: Integer;
begin
  Result := False;
  for i := 0 to Count - 1 do
  begin
    if Dest.IndexOfName(KeyByIndex[i].Name) < 0 then
      Exit;
    if Dest.Values[KeyByIndex[i].Name] <> KeyByIndex[i].Text then
      Exit;
  end;
  Result := True;
end;

procedure TXmlIniSection.CopyFrom(Source: TStrings);
begin
  Clear;
  Append(Source);
end;

procedure TXmlIniSection.Append(Source: TStrings);
var
  i: Integer;
begin
  for i := 0 to Source.Count - 1 do
    SetItem(Source.Names[i], Source.Values[Source.Names[i]]);
end;

function TXmlIniSection.Exists(KeyName: WideString): Boolean;
var
  Index: Integer;
begin
  Result := InternalFind(KeyName, Index);
end;

procedure TXmlIniSection.Clear;
var
  i: Integer;
begin
  FXmlEncoding := xetUTF_16;

  for i := Count - 1 downto 0 do
    KeyByIndex[i].Free;
end;

procedure TXmlIniSection.Collation;
var
  i: Integer;
begin
  for i := Count - 1 downto 0 do
    if KeyByIndex[i].Text = '' then
      KeyByIndex[i].Free;
end;

procedure TXmlIniSection.LoadFromStream(Stream: TStream; SectionName: WideString = '');
var
  XmlIni: TXmlIni;
begin
  XmlIni := TXmlIni.Create;
  try
    XmlIni.LoadFromStream(Stream);

    if (SectionName = '') and not XmlIni.Exists('') then
    begin
      if XmlIni.Exists(FName) then
        SectionName := FName
      else if XmlIni.Count > 0 then
        SectionName := XmlIni.SectionByIndex[0].Name;
    end;

    CopyFrom(XmlIni[SectionName]);
    Name := XmlIni[SectionName].Name;

    FXmlEncoding := XmlIni.XmlEncoding;
    FCompressed := XmlIni.Compressed;
  finally
    XmlIni.Free;
  end;
end;

procedure TXmlIniSection.SaveToStream(Stream: TStream; SectionName: WideString = '');
var
  XmlIni: TXmlIni;
begin
  if SectionName = '' then
    SectionName := FName;

  XmlIni := TXmlIni.Create;
  try
    XmlIni[SectionName].CopyFrom(Self);
    XmlIni.XmlEncoding := FXmlEncoding;
    XmlIni.Compressed := FCompressed;
    XmlIni.SaveToStream(Stream);
  finally
    XmlIni.Free;
  end;
end;

procedure TXmlIniSection.LoadFromString(Value: WideString; SectionName: WideString = '');
var
  SS: TStringStream;
begin
  if TrimLeft(Value) = '' then
    Clear
  else if TrimLeft(Value)[1] = '<' then
  begin
    SS := TStringStream.Create(Value);
    try
      LoadFromStream(SS, SectionName);
    finally
      SS.Free;
    end;
  end
  else
    LoadFromBase64String(Value, SectionName);
end;

function TXmlIniSection.SaveToString(SectionName: WideString = ''): WideString;
var
  SS: TStringStream;
begin
  if not FCompressed then
  begin
    SS := TStringStream.Create('');
    try
      SaveToStream(SS, SectionName);
      Result := SS.DataString;
    finally
      SS.Free;
    end;
  end
  else
    Result := SaveToBase64String;
end;

procedure TXmlIniSection.LoadFromBase64String(Value: string; SectionName: WideString = '');
var
  SS: TStringStream;
  MS: TMemoryStream;
begin
  if (TrimLeft(Value) = '') or (TrimLeft(Value)[1] = '<') then
    LoadFromString(Value)
  else
  begin
    SS := TStringStream.Create(Trim(Value));
    try
      MS := TMemoryStream.Create;
      try
        SS.Position := 0;
        XmlT_MimeDecodeStream(SS, MS);
        LoadFromStream(MS, SectionName);
      finally
        MS.Free;
      end;
    finally
      SS.Free;
    end;
  end;
end;

function TXmlIniSection.SaveToBase64String(SectionName: WideString = ''): string;
var
  SS: TStringStream;
  MS: TMemoryStream;
begin
  MS := TMemoryStream.Create;
  try
    SaveToStream(MS, SectionName);

    SS := TStringStream.Create('');
    try
      MS.Position := 0;
      XmlT_MimeEncodeStream(MS, SS);
      Result := SS.DataString;
    finally
      SS.Free;
    end;
  finally
    MS.Free;
  end;
end;

procedure TXmlIniSection.LoadFromFile(FileName: WideString; SectionName: WideString = '');
var
  XmlIni: TXmlIni;
begin
  XmlIni := TXmlIni.Create;
  try
    XmlIni.LoadFromFile(FileName);

    if (SectionName = '') and not XmlIni.Exists('') then
    begin
      if XmlIni.Exists(FName) then
        SectionName := FName
      else if XmlIni.Count > 0 then
        SectionName := XmlIni.SectionByIndex[0].Name;
    end;

    CopyFrom(XmlIni[SectionName]);
    Name := XmlIni[SectionName].Name;

    FXmlEncoding := XmlIni.XmlEncoding;
    FCompressed := XmlIni.Compressed;
  finally
    XmlIni.Free;
  end;
end;

procedure TXmlIniSection.SaveToFile(FileName: WideString; SectionName: WideString = '');
var
  XmlIni: TXmlIni;
begin
  if SectionName = '' then
    SectionName := FName;

  XmlIni := TXmlIni.Create;
  try
    XmlIni[SectionName].CopyFrom(Self);
    XmlIni.XmlEncoding := FXmlEncoding;
    XmlIni.Compressed := FCompressed;
    XmlIni.SaveToFile(FileName);
  finally
    XmlIni.Free;
  end;
end;

procedure TXmlIniSection.UpdateToFile(FileName: WideString; SectionName: WideString = '');
var
  XmlIni: TXmlIni;
begin
  if SectionName = '' then
    SectionName := FName;

  XmlIni := TXmlIni.Create(FileName);
  try
    XmlIni[SectionName].CopyFrom(Self);
    XmlIni.XmlEncoding := FXmlEncoding;
    XmlIni.Compressed := FCompressed;
    XmlIni.UpdateToFile;
  finally
    XmlIni.Free;
  end;
end;

procedure TXmlIniSection.LoadFromRegistry(Path: WideString; RootKey: HKEY = HKEY_LOCAL_MACHINE; SectionName: WideString = '');
var
  Reg: TRegistry;
  RegDataInfo: TRegDataInfo;
  Names: TStringList;
  i: Integer;
  Buf: PChar;
begin
  Clear;

  if (Path = '') or (Path[Length(Path)] <> '\') then
    Path := Path + '\';

  if SectionName = '' then
    SectionName := FName;

  Reg := TRegistry.Create;
  try
    Reg.RootKey := RootKey;

    if SectionName <> '' then
      Path := Path + SectionName;

    Names := TStringList.Create;
    try
      if Reg.OpenKey(Path, False) then
      begin
        Reg.GetValueNames(Names);

        if Names.Count = 0 then
        begin
          Reg.CloseKey;
          if Reg.OpenKey(Path, False) then
            Reg.GetValueNames(Names);
        end;

        for i := 0 to Names.Count - 1 do
        begin
          Reg.GetDataInfo(Names[i], RegDataInfo);

          case RegDataInfo.RegData of
            rdString, rdExpandString:
              Key[Names[i]].Text := Reg.ReadString(Names[i]);
            rdInteger:
              Key[Names[i]].AsInteger := Reg.ReadInteger(Names[i]);
            rdUnknown, rdBinary:
              begin
                Buf := AllocMem(RegDataInfo.DataSize);
                try
                  Reg.ReadBinaryData(Names[i], Buf^, RegDataInfo.DataSize);
                  Key[Names[i]].FStream.Write(Buf^, RegDataInfo.DataSize);
                finally
                  FreeMem(Buf);
                end;
              end;
          end;
        end;

        Reg.CloseKey;
      end;

      Collation;
    finally
      Names.Free;
    end;
  finally
    Reg.Free;
  end;
end;

procedure TXmlIniSection.SaveToRegistry(Path: WideString; RootKey: HKEY = HKEY_LOCAL_MACHINE; SectionName: WideString = '';
  UseOldDataType: Boolean = False);
var
  Reg: TRegistry;
  RegDataInfo: TRegDataInfo;
  Names: TStringList;
  i: Integer;
begin
  Collation;

  if (Path = '') or (Path[Length(Path)] <> '\') then
    Path := Path + '\';

  if SectionName = '' then
    SectionName := FName;

  Reg := TRegistry.Create;
  try
    Reg.RootKey := RootKey;

    if SectionName <> '' then
      Path := Path + SectionName;

    if Reg.OpenKey(Path, True) then
    begin
      Names := TStringList.Create;
      try
        Reg.GetValueNames(Names);

        for i := 0 to Names.Count - 1 do
          if not UseOldDataType or not Exists(Names[i]) then
            Reg.DeleteValue(Names[i]);
      finally
        Names.Free;
      end;

      for i := 0 to Count - 1 do
      begin
        if UseOldDataType and Reg.ValueExists(KeyByIndex[i].Name) then
        begin
          Reg.GetDataInfo(KeyByIndex[i].Name, RegDataInfo);

          case RegDataInfo.RegData of
            rdString:
              Reg.WriteString(KeyByIndex[i].Name, KeyByIndex[i].Text);
            rdExpandString:
              Reg.WriteExpandString(KeyByIndex[i].Name, KeyByIndex[i].Text);
            rdInteger:
              Reg.WriteInteger(KeyByIndex[i].Name, KeyByIndex[i].AsInteger);
            rdBinary, rdUnknown:
              begin
                KeyByIndex[i].TextToBinary;
                Reg.WriteBinaryData(KeyByIndex[i].Name, KeyByIndex[i].FStream.Memory^, KeyByIndex[i].FStream.Size);
              end;
          else
          end;
        end
        else
        begin
          case KeyByIndex[i].DataType of
            xidtText, xidtInteger, xidtFloat, xidtBool, xidtDateTime, xidtDate, xidtTime:
              begin
                if Reg.ValueExists(KeyByIndex[i].Name) then
                begin
                  Reg.GetDataInfo(KeyByIndex[i].Name, RegDataInfo);
                  if RegDataInfo.RegData <> rdString then
                    Reg.DeleteValue(KeyByIndex[i].Name);
                end;
                Reg.WriteString(KeyByIndex[i].Name, KeyByIndex[i].Text);
              end;
            xidtBinary, xidtUnknow:
              begin
                if Reg.ValueExists(KeyByIndex[i].Name) then
                begin
                  Reg.GetDataInfo(KeyByIndex[i].Name, RegDataInfo);
                  if RegDataInfo.RegData <> rdBinary then
                    Reg.DeleteValue(KeyByIndex[i].Name);
                end;
                KeyByIndex[i].TextToBinary;
                Reg.WriteBinaryData(KeyByIndex[i].Name, KeyByIndex[i].FStream.Memory^, KeyByIndex[i].FStream.Size);
              end;
          end;
        end;
      end;

      Reg.CloseKey;
    end;
  finally
    Reg.Free;
  end;
end;

procedure TXmlIniSection.UpdateToRegistry(Path: WideString; RootKey: HKEY = HKEY_LOCAL_MACHINE; SectionName: WideString = '';
  UseOldDataType: Boolean = False);
var
  Reg: TRegistry;
  RegDataInfo: TRegDataInfo;
  i: Integer;
begin
  Collation;

  if (Path = '') or (Path[Length(Path)] <> '\') then
    Path := Path + '\';

  if SectionName = '' then
    SectionName := FName;

  Reg := TRegistry.Create;
  try
    Reg.RootKey := RootKey;

    if SectionName <> '' then
      Path := Path + SectionName;

    if Reg.OpenKey(Path, True) then
    begin
      for i := 0 to Count - 1 do
      begin
        if UseOldDataType and Reg.ValueExists(KeyByIndex[i].Name) then
        begin
          Reg.GetDataInfo(KeyByIndex[i].Name, RegDataInfo);

          case RegDataInfo.RegData of
            rdString:
              Reg.WriteString(KeyByIndex[i].Name, KeyByIndex[i].Text);
            rdExpandString:
              Reg.WriteExpandString(KeyByIndex[i].Name, KeyByIndex[i].Text);
            rdInteger:
              Reg.WriteInteger(KeyByIndex[i].Name, KeyByIndex[i].AsInteger);
            rdBinary, rdUnknown:
              begin
                KeyByIndex[i].TextToBinary;
                Reg.WriteBinaryData(KeyByIndex[i].Name, KeyByIndex[i].FStream.Memory^, KeyByIndex[i].FStream.Size);
              end;
          else
          end;
        end
        else
        begin
          case KeyByIndex[i].DataType of
            xidtText, xidtInteger, xidtFloat, xidtBool, xidtDateTime, xidtDate, xidtTime, xidtUnknow:
              begin
                if Reg.ValueExists(KeyByIndex[i].Name) then
                begin
                  Reg.GetDataInfo(KeyByIndex[i].Name, RegDataInfo);
                  if RegDataInfo.RegData <> rdString then
                    Reg.DeleteValue(KeyByIndex[i].Name);
                end;
                Reg.WriteString(KeyByIndex[i].Name, KeyByIndex[i].Text);
              end;
            xidtBinary:
              begin
                if Reg.ValueExists(KeyByIndex[i].Name) then
                begin
                  Reg.GetDataInfo(KeyByIndex[i].Name, RegDataInfo);
                  if RegDataInfo.RegData <> rdBinary then
                    Reg.DeleteValue(KeyByIndex[i].Name);
                end;
                KeyByIndex[i].TextToBinary;
                Reg.WriteBinaryData(KeyByIndex[i].Name, KeyByIndex[i].FStream.Memory^, KeyByIndex[i].FStream.Size);
              end;
          end;
        end;
      end;

      Reg.CloseKey;
    end;
  finally
    Reg.Free;
  end;
end;

procedure TXmlIniSection.LoadFromIni(FileName, SectionName: WideString);
var
  XmlIni: TXmlIni;
begin
  XmlIni := TXmlIni.Create;
  try
    XmlIni.LoadFromIni(FileName);

    if (SectionName = '') and not XmlIni.Exists('') then
    begin
      if XmlIni.Exists(FName) then
        SectionName := FName
      else if XmlIni.Count > 0 then
        SectionName := XmlIni.SectionByIndex[0].Name;
    end;

    CopyFrom(XmlIni[SectionName]);
    Name := XmlIni[SectionName].Name;

    FXmlEncoding := XmlIni.XmlEncoding;
    FCompressed := XmlIni.Compressed;
  finally
    XmlIni.Free;
  end;
end;

procedure TXmlIniSection.SaveToIni(FileName, SectionName: WideString);
var
  XmlIni: TXmlIni;
begin
  if SectionName = '' then
    SectionName := FName;

  XmlIni := TXmlIni.Create;
  try
    XmlIni[SectionName].CopyFrom(Self);
    XmlIni.XmlEncoding := FXmlEncoding;
    XmlIni.Compressed := FCompressed;
    XmlIni.SaveToFile(FileName);
  finally
    XmlIni.Free;
  end;
end;

procedure TXmlIniSection.RegisterKey(Key: TXmlIniKey);
var
  Index: Integer;
begin
  Index := FList.IndexOf(Key);
  if Index < 0 then
    FList.Add(Key);

  if InternalFind(Key.FName, Index) then
    FSortedList.Delete(Index);

  InternalFind(Key.FName, Index);
  FSortedList.Insert(Index, Key);
end;

procedure TXmlIniSection.UnregisterKey(Key: TXmlIniKey);
var
  Index: Integer;
begin
  Index := FList.IndexOf(Key);
  if Index >= 0 then
    FList.Delete(Index);

  if InternalFind(Key.FName, Index) then
    FSortedList.Delete(Index);
end;

function TXmlIniSection.InternalFindCaseSensitive(const S: WideString; var Index: Integer): Boolean;
var
  L, H, I, C: Integer;
begin
  Result := False;
  L := 0;
  H := FSortedList.Count - 1;
  while L <= H do
  begin
    I := (L + H) shr 1;
    C := WideCompareStr(TXmlIniKey(FSortedList[I]).FName, S);
    if C < 0 then L := I + 1 else
    begin
      H := I - 1;
      if C = 0 then
      begin
        Result := True;
        L := I;
      end;
    end;
  end;
  Index := L;
end;

function TXmlIniSection.InternalFindCaseInsensitive(const S: WideString; var Index: Integer): Boolean;
var
  L, H, I, C: Integer;
begin
  Result := False;
  L := 0;
  H := FSortedList.Count - 1;
  while L <= H do
  begin
    I := (L + H) shr 1;
    C := WideCompareText(TXmlIniKey(FSortedList[I]).FName, S);
    if C < 0 then L := I + 1 else
    begin
      H := I - 1;
      if C = 0 then
      begin
        Result := True;
        L := I;
      end;
    end;
  end;
  Index := L;
end;

function TXmlIniSection.InternalFind(const KeyName: WideString; var Index: Integer): Boolean;
begin
  if FCaseSensitive then
    Result := InternalFindCaseSensitive(Trim(KeyName), Index)
  else
    Result := InternalFindCaseInsensitive(Trim(KeyName), Index);
end;

function TXmlIniSection.InternalAdd(KeyName: WideString): TXmlIniKey;
begin
  Result := TXmlIniKey.Create(KeyName, '', Self);
end;

procedure TXmlIniSection.SetCaseSensitive(Value: Boolean);
var
  i: Integer;
begin
  if FCaseSensitive <> Value then
  begin
    FCaseSensitive := Value;

    FSortedList.Clear;
    for i := 0 to Count - 1 do
      RegisterKey(KeyByIndex[i]);
  end;
end;

procedure TXmlIniSection.SetName(Value: WideString);
begin
  try
    Value := Trim(Value);
    if Value <> FName then
    begin
      if FOwner <> nil then
      begin
        if FOwner.Exists(Value) then
          raise Exception.Create('Section "' + Value + '" already exists.');
        FOwner.RegisterSection(Self);
      end;
      FName := Value;
    end;
  except
    if FRaiseException then
      raise;
  end;
end;

function TXmlIniSection.GetKeyByIndex(Index: Integer): TXmlIniKey;
begin
  if (Index >= 0) and (Index < FList.Count) then
    Result := TXmlIniKey(FList[Index])
  else
    Result := GetKey('');
end;

function TXmlIniSection.GetKey(KeyName: WideString): TXmlIniKey;
var
  Index: Integer;
begin
  if InternalFind(KeyName, Index) then
    Result := TXmlIniKey(FSortedList[Index])
  else
    Result := InternalAdd(KeyName);
end;

function TXmlIniSection.GetItem(Index: Variant): WideString;
var
  IntIndex: Integer;
begin
  if VarIsNumeric(Index) then
  begin
    IntIndex := StrToIntDef(VarToWideStr(Index), 0);
    if (IntIndex >= 0) and (IntIndex < Count) then
      Result := KeyByIndex[IntIndex].Text
    else
      Result := '';
  end
  else
  begin
    if InternalFind(VarToWideStr(Index), IntIndex) then
      Result := TXmlIniKey(FSortedList[IntIndex]).Text
    else
      Result := '';
  end;
end;

procedure TXmlIniSection.SetItem(Index: Variant; const Value: WideString);
begin
  if VarIsNumeric(Index) then
    KeyByIndex[StrToIntDef(VarToWideStr(Index), 0)].Text := Value
  else
    Key[VarToWideStr(Index)].Text := Value;
end;

function TXmlIniSection.GetCount: Integer;
begin
  Result := FList.Count;
end;

function TXmlIniSection.GetIndex: Integer;
begin
  if FOwner <> nil then
    Result := FOwner.FList.IndexOf(Self)
  else
    Result := -1;
end;

procedure TXmlIniSection.SetRaiseException(const Value: Boolean);
var
  i: Integer;
begin
  if FRaiseException <> Value then
  begin
    FRaiseException := Value;
    for i := 0 to Count - 1 do
      KeyByIndex[i].RaiseException := FRaiseException;
  end;
end;

{ TXmlIniKey }

constructor TXmlIniKey.Create(AName: WideString = ''; AText: WideString = ''; AOwner: TXmlIniSection = nil);
begin
  FOwner := AOwner;

  if FOwner <> nil then
    FRaiseException := FOwner.RaiseException
  else
    FRaiseException := False;

  FName := Trim(AName);
  FStream := TXmlIniEventMemoryStream.Create(Self);
  FDataType := xidtText;
  Text := AText;

  if FOwner <> nil then
    FOwner.RegisterKey(Self);
end;

destructor TXmlIniKey.Destroy;
begin
  if FOwner <> nil then
    FOwner.UnregisterKey(Self);

  FStream.Free;

  inherited;
end;

procedure TXmlIniKey.Clear;
begin
  FText := '';
  FStream.Clear;
  FDataType := xidtText;
end;

procedure TXmlIniKey.ValueLoadFromFile(FileName: WideString; Binary: Boolean = True);
var
  SS: TStringStream;
  MS: TMemoryStream;
begin
  try
    Clear;

    FileName := Trim(FileName);

    if (FileName <> '') and (ExtractFilePath(FileName) = '') then
      FileName := ExtractFilePath(ParamStr(0)) + FileName;

    if (FileName = '') or not FileExists(FileName) then
      if FRaiseException then
        raise Exception.Create('File "' + FileName + '" not exists.')
      else
        Exit;

    if Binary then
      FStream.LoadFromFile(FileName)
    else
    begin
      SS := TStringStream.Create('');
      try
        MS := TMemoryStream.Create;
        try
          MS.LoadFromFile(FileName);
          SS.CopyFrom(MS, 0);
        finally
          MS.Free;
        end;
        Text := SS.DataString;
      finally
        SS.Free;
      end;
    end;
  except
    if FRaiseException then
      raise;
  end;
end;

procedure TXmlIniKey.ValueSaveToFile(FileName: WideString);
var
  SS: TStringStream;
  MS: TMemoryStream;
begin
  FileName := Trim(FileName);

  if (FileName <> '') and (ExtractFilePath(FileName) = '') then
    FileName := ExtractFilePath(ParamStr(0)) + FileName;

  if FileName = '' then
    if FRaiseException then
      raise Exception.Create('File name is required.')
    else
      Exit;

  try
    if DataType = xidtBinary then
    begin
      if FStream.Size > 0 then
        FStream.SaveToFile(FileName);
    end
    else
    begin
      if FText <> '' then
      begin
        SS := TStringStream.Create(FText);
        try
          MS := TMemoryStream.Create;
          try
            MS.CopyFrom(SS, 0);
            MS.SaveToFile(FileName);
          finally
            MS.Free;
          end;
        finally
          SS.Free;
        end;
      end;
    end;
  except
    if FRaiseException then
      raise;
  end;
end;

procedure TXmlIniKey.SetName(Value: WideString);
begin
  try
    Value := Trim(Value);
    if Value <> FName then
    begin
      if FOwner <> nil then
      begin
        if FOwner.Exists(Value) then
          raise Exception.Create('Key "' + Value + '" already exists.');
        FOwner.RegisterKey(Self);
      end;
      FName := Value;
    end;
  except
    if FRaiseException then
      raise;
  end;
end;

procedure TXmlIniKey.BinaryToText;
var
  SS: TStringStream;
begin
  try
    if DataType = xidtBinary then
    begin
      SS := TStringStream.Create('');
      try
        FStream.Position := 0;
        XmlT_MimeEncodeStream(FStream, SS);
        FText := XI_Base64Header + SS.DataString;
      finally
        SS.Free;
      end;
    end
    else
    begin
      SS := TStringStream.Create('');
      try
        SS.CopyFrom(FStream, 0);
        FText := SS.DataString;
      finally
        SS.Free;
      end;
    end;
  except
    if FRaiseException then
      raise;
  end;
end;

procedure TXmlIniKey.TextToBinary;
var
  SS: TStringStream;
  Base64Text: WideString;
  OldDataType: TXmlIniDataType;
begin
  try
    if Copy(FText, 1, Length(XI_Base64Header)) <> XI_Base64Header then
    begin
      SS := TStringStream.Create(FText);
      try
        OldDataType := DataType;
        FStream.CopyFrom(SS, 0);
        FDataType := OldDataType;
      finally
        SS.Free;
      end;
    end
    else
    begin
      Base64Text := Trim(FText);

      Delete(Base64Text, 1, 8);

      SS := TStringStream.Create(Base64Text);
      try
        XmlT_MimeDecodeStream(SS, FStream);
      finally
        SS.Free;
      end;
    end;
  except
    if FRaiseException then
      raise;
  end;
end;

function TXmlIniKey.GetText: WideString;
begin
  if (FText = '') and (FStream.Size > 0) then
    BinaryToText;

  Result := FText;
end;

procedure TXmlIniKey.SetText(const Value: WideString);
begin
  if FText <> Value then
  begin
    FStream.Clear;

    FText := Value;
    FDataType := xidtUnknow;
  end;
end;

function TXmlIniKey.GetAsStream: TStream;
begin
  Result := FStream;
end;

procedure TXmlIniKey.SetAsStream(const Value: TStream);
begin
  FText := '';

  if (Value <> nil) and (Value.Size > 0) then
  begin
    Value.Position := 0;
    FStream.CopyFrom(Value, 0);
    FDataType := xidtBinary;
  end
  else
  begin
    FStream.Clear;
    FDataType := xidtText;
  end;
end;

function TXmlIniKey.GetAsInteger: Integer;
begin
  Result := StrToIntDef(Text, 0);
end;

procedure TXmlIniKey.SetAsInteger(const Value: Integer);
begin
  Text := IntToStr(Value);
  FDataType := xidtInteger;
end;

function TXmlIniKey.GetAsFloat: Double;
begin
  Result := StrToFloatDef(Text, 0);
end;

procedure TXmlIniKey.SetAsFloat(const Value: Double);
begin
  Text := FloatToStr(Value);
  FDataType := xidtFloat;
end;

function TXmlIniKey.GetAsBool: Boolean;
begin
  Result := XmlT_StrToBool(Text);
end;

procedure TXmlIniKey.SetAsBool(const Value: Boolean);
begin
  Text := XmlT_BoolToStr(Value);
  FDataType := xidtBool;
end;

function TXmlIniKey.GetAsDateTime: TDateTime;
begin
  Result := XmlT_StrToDateTime(Text);
end;

procedure TXmlIniKey.SetAsDateTime(const Value: TDateTime);
begin
  Text := XmlT_DateTimeToStr(Value);
  FDataType := xidtDateTime;
end;

function TXmlIniKey.GetAsDate: TDateTime;
begin
  Result := XmlT_StrToDate(Text);
end;

procedure TXmlIniKey.SetAsDate(const Value: TDateTime);
begin
  Text := XmlT_DateToStr(Value);
  FDataType := xidtDate;
end;

function TXmlIniKey.GetAsTime: TDateTime;
begin
  Result := XmlT_StrToTime(Text);
end;

procedure TXmlIniKey.SetAsTime(const Value: TDateTime);
begin
  Text := XmlT_TimeToStr(Value);
  FDataType := xidtTime;
end;

function TXmlIniKey.GetDataType: TXmlIniDataType;
begin
  if FDataType = xidtUnknow then
    FDataType := DetectDataType(FText);
  Result := FDataType;
end;

procedure TXmlIniKey.SetDataType(const Value: TXmlIniDataType);
var
  SS: TStringStream;
begin
  if DataType <> Value then
  begin
    if Value = xidtBinary then
    begin
      FStream.Clear;

      try
        SS := TStringStream.Create(FText);
        try
          FStream.CopyFrom(SS, 0);
        finally
          SS.Free;
        end;
      except
        if FRaiseException then
          raise;
      end;
    end
    else if DataType = xidtBinary then
    begin
      try
        if FStream.Size = 0 then
          FText := ''
        else
        begin
          SS := TStringStream.Create('');
          try
            SS.CopyFrom(FStream, 0);
            FText := SS.DataString;
          finally
            SS.Free;
          end;
        end;
        FStream.Clear;
      except
        if FRaiseException then
          raise;
      end;
    end;

    FDataType := Value;

    case FDataType of
      xidtText:
        Text := Text;
      xidtInteger:
        AsInteger := AsInteger;
      xidtFloat:
        AsFloat := AsFloat;
      xidtBool:
        AsBool := AsBool;
      xidtDateTime:
        AsDateTime := AsDateTime;
      xidtDate:
        AsDate := AsDate;
      xidtTime:
        AsTime := AsTime;
    end;
  end;
end;

function TXmlIniKey.GetIndex: Integer;
begin
  if FOwner <> nil then
    Result := FOwner.FList.IndexOf(Self)
  else
    Result := -1;
end;

function TXmlIniKey.DetectDataType(Data: WideString): TXmlIniDataType;
  function InArray(Str: WideString; ArrayStr: array of WideString): Boolean;
  var
    i: Integer;
  begin
    Result := False;

    Str := UpperCase(Str);

    for i := Low(ArrayStr) to High(ArrayStr) do
      if Str = UpperCase(ArrayStr[i]) then
      begin
        Result := True;
        Break;
      end;
  end;
  function IsInteger(Str: WideString): Boolean;
  begin
    Result := Str = IntToStr(StrToIntDef(Str, 0));
  end;
  function IsFloat(Str: WideString): Boolean;
  begin
    Result := Str = FloatToStr(StrToFloatDef(Str, 0));
  end;
  function IsDateTime(Str: WideString): Boolean;
  begin
    Result := Str = XmlT_DateTimeToStr(XmlT_StrToDateTime(Str));
  end;
  function IsDate(Str: WideString): Boolean;
  begin
    Result := Str = XmlT_DateToStr(XmlT_StrToDate(Str));
  end;
  function IsTime(Str: WideString): Boolean;
  begin
    Result := Str = XmlT_DateToStr(XmlT_StrToTime(Str));
  end;
begin
  if (Data <> '') and (Copy(Data, 1, Length(XI_Base64Header)) = XI_Base64Header) then
    Result := xidtBinary
  else if IsInteger(Data) then
    Result := xidtInteger
  else if InArray(Data, XT_True_Text) or InArray(Data, XT_False_Text) then
    Result := xidtBool
  else if IsFloat(Data) then
    Result := xidtFloat
  else if IsDateTime(Data) then
    Result := xidtDateTime
  else if IsDate(Data) then
    Result := xidtDate
  else if IsTime(Data) then
    Result := xidtTime
  else
    Result := xidtText;
end;

{ TXmlIniEventMemoryStream }

constructor TXmlIniEventMemoryStream.Create(AOwner: TXmlIniKey);
begin
  FOwner := AOwner;

  inherited Create;
end;

function TXmlIniEventMemoryStream.GetSize: Int64;
begin
  if (inherited GetSize = 0) and (FOwner.FText <> '') then
    FOwner.TextToBinary;

  Result := inherited GetSize;
end;

procedure TXmlIniEventMemoryStream.SetSize(NewSize: Integer);
begin
  FOwner.FDataType := xidtBinary;
  FOwner.FText := '';

  inherited SetSize(NewSize);
end;

function TXmlIniEventMemoryStream.Read(var Buffer; Count: Integer): Longint;
begin
  if (inherited GetSize = 0) and (FOwner.FText <> '') then
    FOwner.TextToBinary;

  Result := inherited Read(Buffer, Count);
end;

function TXmlIniEventMemoryStream.Write(const Buffer; Count: Integer): Longint;
begin
  FOwner.FDataType := xidtBinary;
  FOwner.FText := '';

  Result := inherited Write(Buffer, Count);
end;

end.

